# -*- coding: utf8 -*-
import base64
import hashlib
import json
import logging
import os
import random
import socket
import sys
import types

import click
from pandas.io.json import json_normalize
from requests import HTTPError
from terminaltables import PorcelainTable
from six.moves.urllib import parse

from mali_commands.legit.api import BASE_URL_PATH
from .utilities.tables import format_json_data, dict_to_csv


def print_as_json(result):
    if isinstance(result, types.GeneratorType):
        try:
            first_result = next(result)
        except StopIteration:
            sys.stdout.write('[]')
            return

        is_list = len(first_result) == 1 or isinstance(first_result, dict)

        if is_list:
            sys.stdout.write('[')
            json.dump(first_result, sys.stdout)

            for next_result in result:
                sys.stdout.write(',')
                json.dump(next_result, sys.stdout)

            sys.stdout.write(']')
        else:
            sys.stdout.write('{')

            key, val = first_result

            def write_key_val():
                json.dump(key, sys.stdout)
                sys.stdout.write(':')
                json.dump(val, sys.stdout)

            write_key_val()
            for key, val in result:
                sys.stdout.write(',\n')
                write_key_val()

            sys.stdout.write('}')
    else:
        json.dump(result, sys.stdout)


def create_validate_length(min_length, max_length):
    def validate_length(ctx, option, value):
        if value is None:
            return value

        if len(value) > max_length:
            raise click.BadParameter("needs to be shorter than %s characters" % max_length)

        if len(value) < min_length:
            raise click.BadParameter("needs to be longer than %s characters" % min_length)

        return value

    return validate_length


def add_to_data_if_not_none(data, val, key):
    if val is not None:
        data[key] = val


def get_data_and_remove(kwargs, key):
    key = key.lower()
    result = kwargs.get(key)

    if key in kwargs:
        del kwargs[key]

    return result


def base64url(b):
    return bytearray(base64.b64encode(b).decode('ascii').replace('=', '').replace('+', '-').replace('/', '_'), 'ascii')


def sha256(s):
    h = hashlib.sha256()
    h.update(s)
    return h.digest()


def urljoin(*args):
    base = args[0]
    for u in args[1:]:
        base = parse.urljoin(base, u)

    return base


class LocalWebServer(object):
    ports = [27017, 41589, 55130, 6617, 1566, 17404, 44288, 6948, 15950, 48216, 3318, 34449, 22082, 22845, 37862, 59304]

    def __init__(self):
        self.app = self.create_app()
        self.code = None
        self.srv = None

        self.disable_logging()
        self.start_server()

    def close(self):
        if self.srv is not None:
            self.srv.server_close()

    @classmethod
    def disable_logging(cls):
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    def create_app(self):
        from flask import Flask, request, redirect
        app = Flask(__name__)

        @app.route('/console/auth/init')
        def auth_init():
            self.code = request.args.get('code')
            return redirect("/console/auth/finish")

        @app.route('/console/auth/finish')
        def auth_finish():
            self.srv.shutdown_signal = True
            return "You can go now back to the console"

        return app

    def run(self):
        self.srv.serve_forever()

    @property
    def redirect_uri(self):
        return 'http://localhost:%s/console/auth/init' % self.srv.port

    def start_server(self):
        from werkzeug.serving import make_server

        ports = self.ports[:]
        random.shuffle(ports)

        for port in ports:
            try:
                self.srv = make_server('127.0.0.1', port, self.app)
                break
            except (OSError, socket.error):
                continue

        if self.srv is None:
            click.echo("can't create local server for authentication, consider using the --disable-webserver flag")


class WaitForHttpResponse:
    ports = [27017, 41589, 55130, 6617, 1566, 17404, 44288, 6948, 15950, 48216, 3318, 34449, 22082, 22845, 37862, 59304]

    def __init__(self, on_response=None, uri='/', methods=['GET'], port=None):
        self.uri = uri
        self.methods = methods
        self.on_response = on_response
        self.app = self.create_app()
        self.port = port
        self.response = None
        self.srv = None
        self.url = None
        self.disable_logging()
        self.start_server()

    def close(self):
        if self.srv is not None:
            self.srv.server_close()

    @classmethod
    def disable_logging(cls):
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    def create_app(self):
        from flask import Flask, request
        app = Flask(__name__)

        @app.route(self.uri, methods=self.methods)
        def read_result():
            self.response = dict(request.args)
            result = None
            if self.on_response is not None:
                result = self.on_response(request)
            self.srv.shutdown_signal = True
            return result or 'ok'

        return app

    def run(self):
        self.srv.serve_forever()
        return self.response

    def start_server(self):
        from werkzeug.serving import make_server

        ports = self.ports[:] if self.port is None else [self.port]
        random.shuffle(ports)
        for port in ports:
            try:
                self.srv = make_server('127.0.0.1', port, self.app)
                self.url = 'http://127.0.0.1:%s%s' % (port, self.uri)
                break
            except (OSError, socket.error):
                continue

        if self.srv is None:
            click.echo("can't create local server for authentication, consider using the --disable-webserver flag")


def wait_for_input():
    if sys.version_info >= (3, 0):
        code = input('Enter the token ')
    else:
        # noinspection PyUnresolvedReferences
        code = raw_input('Enter the token ')

    return code


def ml_auth_oauth_token(api_host, session, code, verifier, redirect_uri):
    params = {
        'code': code,
        'code_verifier': verifier,
        'grant_type': 'authorization_code',
        'redirect_uri': redirect_uri,
    }

    url = '{}/{}users/token'.format(api_host, BASE_URL_PATH)
    r = session.post(url, json=params)

    r.raise_for_status()

    return r.json()


# noinspection PyUnusedLocal
def pixy_flow(ctx):
    import webbrowser

    verifier = base64url(os.urandom(32))
    verifier_challenge = base64url(sha256(verifier))
    verifier_challenge = verifier_challenge.decode('ascii')

    verifier = verifier.decode('ascii')

    def get_authorize_url():
        current_redirect_uri = app.redirect_uri if ctx.local_web_server else '{host}/console/auth/init'.format(host=ctx.host)

        query = {
            'redirect_uri': current_redirect_uri,
            'code_challenge': verifier_challenge,
        }

        return '{}/console/auth?{}'.format(ctx.host, parse.urlencode(query)), current_redirect_uri

    app = LocalWebServer() if ctx.local_web_server else None

    authorize_url, redirect_uri = get_authorize_url()

    if ctx.local_web_server and not webbrowser.open(authorize_url):
        app.close()
        ctx.local_web_server = False
        authorize_url, redirect_uri = get_authorize_url()

    if ctx.local_web_server:
        msg = "If the browser doesn't open enter this URL manually\n{authorize_url}\n"
    else:
        msg = "Enter the following URL in your machine browser\n{authorize_url}\n"

    click.echo(msg.format(authorize_url=authorize_url))

    if ctx.local_web_server:
        app.run()
        code = app.code
    else:
        code = wait_for_input()

    try:
        data = ml_auth_oauth_token(ctx.api_host, ctx.session, code, verifier, redirect_uri)
    except HTTPError:
        click.echo("Failed to init authorized (did you enter the correct token?).", err=True)
        sys.exit(1)

    click.echo("Success!, you are authorized to use the command line.")

    return data['access_token'], data['refresh_token'], data['id_token']


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def output_result(ctx, result, fields=None, formatters=None, write_row_index=False):
    if result is None:
        return

    format_tables = ctx.obj.output_format == 'tables'
    format_csv = ctx.obj.output_format == 'csv'

    result = format_json_data(result, formatters)

    def normal_name(name):
        return name.replace('_', ' ').title()

    if format_csv:
        df = json_normalize(list(result))
        df = df.rename(normal_name, axis='columns')

        df.to_csv(sys.stdout, columns=map(normal_name, list(df.columns)), index=write_row_index)
    elif format_tables:
        table_data = list(dict_to_csv(result, fields))

        click.echo(PorcelainTable(table_data).table)
    else:
        print_as_json(result)



# -*- coding: utf8 -*-
import six

from .luqum.tree import OrOperation, Word, Item, Phrase, SearchField, UnknownOperation
from .luqum.utils import LuceneTreeVisitorV2, UnknownOperationResolver, LuceneTreeTransformer
from .luqum.parser import parser, lexer, ParseError
import pyparsing as pp


def process_vals(t, l):
    res = []
    for val in t:
        res.append(l(val))

    return res


def convert_number(t):
    def int_or_float(val):
        if '.' in val:
            return float(val)

        return int(val)

    return process_vals(t, int_or_float)


def convert_boolean(t):
    def bools(val):
        if val.lower() in ['true', 'yes']:
            return True

        return False

    return process_vals(t, bools)


def parse_expr(expr_text):
    operator = pp.Regex(">=|<=|!=|>|<|=")('operator')

    point = pp.Literal('.')
    e = pp.CaselessLiteral('E')
    plus_or_minus = pp.Literal('+') | pp.Literal('-')
    number = pp.Word(pp.nums)
    integer = pp.Combine(pp.Optional(plus_or_minus) + number)
    float_number = pp.Combine(integer + pp.Optional(point + pp.Optional(number)) + pp.Optional(e + integer) + pp.stringEnd)

    float_number.setParseAction(convert_number)

    TRUE = pp.Keyword("True", caseless=True) | pp.Keyword("Yes", caseless=True)
    FALSE = pp.Keyword("False", caseless=True) | pp.Keyword("No", caseless=True)

    boolean = (TRUE | FALSE)("boolean")
    boolean.setParseAction(convert_boolean)
    identifier = ~float_number + pp.Word(pp.alphanums + "_" + "." + "-")

    quoted_string = pp.QuotedString('"')
    text_identifier = (quoted_string | identifier)
    comparison_term = boolean | text_identifier | float_number
    values = pp.OneOrMore(comparison_term)('values')
    range_op = pp.Suppress(pp.CaselessLiteral('TO'))
    range_operator = (pp.Suppress('[') + comparison_term + range_op + comparison_term + pp.Suppress(']'))('range')
    range_inclusive = (pp.Suppress('{') + comparison_term + range_op + comparison_term + pp.Suppress('}'))('range_inclusive')

    condition_group = pp.Group((range_inclusive | range_operator | values | (operator + values)))('condition_group')
    conditions = pp.Group(condition_group)('conditions')

    expr = pp.infixNotation(conditions, [
        ("AND", 2, pp.opAssoc.LEFT,),
        ("OR", 2, pp.opAssoc.LEFT,),
    ])

    return expr.parseString(expr_text)


def __conditions_expr_to_sql_where(field_name, item):
    return '({sql})'.format(sql=expr_to_sql_where(field_name, item))


def quotes_if_needed(val):
    return '"%s"' % val if isinstance(val, six.string_types) else val


def __condition_group_expr_to_sql_where(field_name, item):
    if 'values' in item:
        values = item.get('values')
        operator = item.get('operator', '=')
        vals = []
        for val in values:
            vals.append('`{field_name}`{operator}{value}'.format(
                field_name=field_name, operator=operator, value=quotes_if_needed(val)))

        return 'OR'.join(vals)

    if 'range' in item:
        return ('(`{field_name}`>={value1})AND(`{field_name}`<={value2})'.format(
            field_name=field_name, value1=item[0], value2=item[1]))

    if 'range_inclusive' in item:
        return ('(`{field_name}`>{value1})AND(`{field_name}`<{value2})'.format(
            field_name=field_name, value1=item[0], value2=item[1]))


def expr_to_sql_where(field_name, expr):
    sql = ''
    for item in expr:
        if item.getName() == 'conditions':
            sql += __conditions_expr_to_sql_where(field_name, item)
        elif item.getName() == 'condition_group':
            sql += __condition_group_expr_to_sql_where(field_name, item)

    return sql


class QueryFunction(Item):
    def __init__(self, name, node):
        self.name = name
        self.node = node

    def __str__(self):
        return str(self.node)

    @classmethod
    def create_with_args(cls, vals):
        name = cls.__name__.replace('Function', '').lower()
        node = SearchField('@' + name, vals)
        return cls(name, node)


class FunctionVersion(QueryFunction):
    def __init__(self, name, node):
        super(FunctionVersion, self).__init__(name, node)
        self.version = str(node.expr)


class FunctionPhase(QueryFunction):
    def __init__(self, name, node):
        super(FunctionPhase, self).__init__(name, node)
        self.phase = str(node.expr)


class FunctionSplit(QueryFunction):
    def __init__(self, name, node):
        super(FunctionSplit, self).__init__(name, node)

        self.split_field = None
        self.split = {}

        split_vars = self.__validate_split_params(node)

        try:
            split_vars = list(map(float, split_vars))
            split_vars += [None] * (3 - len(split_vars))

            self.split = dict(zip(('train', 'test', 'validation'), split_vars))
        except ValueError:
            if len(split_vars) == 1:
                self.split_field = split_vars[0]
            else:
                raise ParseError('invalid values in @split')

    @classmethod
    def __validate_split_params(cls, node):
        split_vars = str(node)
        split_vars = split_vars.split(':')

        if len(split_vars) > 4:
            raise ParseError('too many values in @split')

        if len(split_vars) == 1:
            raise ParseError('@split needs at least one value')

        return split_vars[1:]


class FunctionSample(QueryFunction):
    def __init__(self, name, node):
        super(FunctionSample, self).__init__(name, node)
        self.sample = float(node.expr.value)


class FunctionSeed(QueryFunction):
    def __init__(self, name, node):
        super(FunctionSeed, self).__init__(name, node)
        self.seed = int(str(node.expr))


class FunctionGroupBy(QueryFunction):
    def __init__(self, name, node):
        super(FunctionGroupBy, self).__init__(name, node)
        self.group = str(node.expr)


class FunctionDatapointBy(QueryFunction):
    def __init__(self, name, node):
        super(FunctionDatapointBy, self).__init__(name, node)
        self.datapoint = str(node.expr)


FunctionGroup = FunctionGroupBy


class FunctionLimit(QueryFunction):
    def __init__(self, name, node):
        super(FunctionLimit, self).__init__(name, node)
        self.limit = int(str(node.expr))


class FunctionOffset(QueryFunction):
    def __init__(self, name, node):
        super(FunctionOffset, self).__init__(name, node)
        self.offset = int(str(node.expr))


class MLQueryMixin(object):
    @classmethod
    def _handle_internal_function(cls, node):
        if node.name.startswith('@'):
            func_name = node.name[1:]
            func_class_name = 'Function%s' % func_name.title()
            func_class_name = func_class_name.replace('_', '')
            function_class = globals().get(func_class_name)

            if function_class is None:
                raise ParseError('invalid function %s' % func_name)

            return function_class(func_name, node)

        return None


# noinspection PyClassicStyleClass
class MLQueryTransformer(LuceneTreeTransformer, MLQueryMixin):
    def visit_search_field(self, node, parents):
        return self._handle_internal_function(node) or node

    def __call__(self, tree):
        return self.visit(tree)


class MLQueryVisitor(LuceneTreeVisitorV2):
    def __visit_binary_operation(self, node, parents, context, op):
        new_context = {}
        if context is not None:
            new_context.update(context)

        new_context['op'] = op
        for child in node.children:
            self.visit(child, context=new_context)

    def visit_and_operation(self, node, parents=None, context=None):
        self.__visit_binary_operation(node, parents, context, 'AND')

    def visit_or_operation(self, node, parents=None, context=None):
        self.__visit_binary_operation(node, parents, context, 'OR')


# noinspection PyClassicStyleClass
class SQLQueryBuilder(MLQueryVisitor, MLQueryMixin):
    def __init__(self, sql_helper):
        self.__where = {}
        self.__limit = []
        self.__vars = {}
        self.__sql_helper = sql_helper

    internal_function_names = ['sample', 'seed', 'group', 'group_by', 'datapoint_by', 'offset', 'version', 'limit']

    def visit_function_sample(self, node, parents=None, context=None):
        sample_percentile = 1.0 - node.sample

        self.__where.setdefault(1, []).append(
            ('AND', '($random_function>{sample_percentile:.4g})'.format(sample_percentile=sample_percentile)))
        self.__vars['sample_percentile'] = sample_percentile
        self.__vars['sample'] = node.sample

    def visit_function_seed(self, node, parents=None, context=None):
        self.__vars['seed'] = node.seed

    def visit_function_limit(self, node, parents=None, context=None):
        self.__vars['limit'] = node.limit

    def visit_function_offset(self, node, parents=None, context=None):
        self.__vars['offset'] = node.offset

    def visit_function_group(self, node, parents=None, context=None):
        self.visit_function_group_by(node, parents, context)

    def visit_function_group_by(self, node, parents=None, context=None):
        self.__vars['group'] = node.group

    def visit_function_datapoint_by(self, node, parents=None, context=None):
        self.__vars['datapoint'] = node.datapoint

    def visit_function_split(self, node, parents=None, context=None):
        if node.split_field is not None:
            self.__vars['split_field'] = node.split_field
        else:
            self.__vars.update(get_split_vars(node.split))

    def visit_function_phase(self, node, parents=None, context=None):
        self.__add_where('_phase', node.phase, '=')

    def visit_function_version(self, node, parents=None, context=None):
        self.__vars['version'] = node.version

    def visit_func_limit(self, node, parents=None, context=None):
        self.__vars['limit'] = node.limit

    def visit_group(self, node, parents=None, context=None):
        self._sub_visit(node.expr, context=context, on_sql=self.wrap_in_parentheses)

    def visit_not(self, node, parents=None, context=None):
        def wrap_in_not(sql):
            return 'NOT ' + self.wrap_in_parentheses(sql)

        self._sub_visit(node.a, context=context, on_sql=wrap_in_not)

    def visit_prohibit(self, node, parents=None, context=None):
        self.__add_where(context.get('field_name'), str(node), context.get('op'))

    def visit_phrase(self, node, parents=None, context=None):
        phrase = str(node)

        phrase = phrase.strip()

        self.__add_where(context.get('field_name'), phrase, context.get('op'))

    def visit_range(self, node, parents=None, context=None):
        self.__add_where(context.get('field_name'), str(node), context.get('op'))

    def visit_plus(self, node, parents=None, context=None):
        self.__add_where(context.get('field_name'), str(node)[1:], context.get('op'))

    def visit_field_group(self, node, parents=None, context=None):
        self.visit(node.expr, parents + [node], context)

    def __add_where(self, field_name, expr, op):
        if expr is None:
            self.__where.setdefault(0, []).append((None, '%s is NULL' % field_name))
            return

        where = expr_to_sql_where(field_name, parse_expr(expr))
        self.__where.setdefault(0, []).append((op, where))

    def visit_word(self, node, parents=None, context=None):
        if context is None:
            raise ParseError('invalid field %s' % node)

        self.__add_where(context.get('field_name'), str(node), context.get('op'))

    def visit_search_field(self, node, parents=None, context=None):
        function_node = self._handle_internal_function(node)
        if function_node is not None:
            return function_node

        context = context or {}

        for child in node.children:
            if isinstance(child, (Word, Phrase)):
                value = str(node.expr)
                if value.lower() == 'null':
                    self.__add_where(node.name, None, 'is')
                else:
                    self.__add_where(node.name, value, context.get('op'))
            else:
                sub_context = {}
                sub_context.update(context)
                sub_context['field_name'] = node.name
                self._sub_visit(node.expr, context=sub_context)

    @classmethod
    def __combine_where(cls, sql_builder):
        combine_where = {}
        for bucket, wheres in sorted(sql_builder.where.items()):
            for where in wheres:
                operator, expr = where

                combine_where.setdefault(bucket, []).append((operator, expr))

        return combine_where

    def _sub_visit(self, child, context=None, on_sql=None):
        sql_builder = SQLQueryBuilder(self.__sql_helper)
        sql_builder.visit(child, context=context)

        self.__vars.update(sql_builder.vars)

        combine_where = self.__combine_where(sql_builder)

        for bucket, wheres in combine_where.items():
            op = None
            sql = ''
            for where in wheres:
                operator, expr = where
                if op is not None:
                    sql += op

                sql += expr
                op = operator

            if on_sql:
                sql = on_sql(sql)
            elif len(wheres) > 1:
                sql = self.wrap_in_parentheses(sql)

            op = (context or {}).get('op')
            self.__where.setdefault(bucket, []).append((op, sql))

    @classmethod
    def wrap_in_parentheses(cls, text):
        return '(%s)' % text

    def build_vars(self):
        if not self.__vars:
            return None

        return self.__vars

    def __build_where(self):
        sql = ''
        for bucket, wheres in sorted(self.__where.items()):
            last_operator = None

            for where in wheres:
                operator, expr = where

                if last_operator:
                    sql += last_operator
                elif sql:
                    sql += operator

                sql += expr

                last_operator = operator

        return sql

    def build_where(self):
        return self.__build_where() or None  # Make sure to return None and not empty string

    @property
    def where(self):
        return self.__where

    @property
    def vars(self):
        return self.__vars


# noinspection PyClassicStyleClass
class AddLimitOffsetFunctionTransformer(LuceneTreeTransformer):
    def __init__(self, limit, offset):
        self.__offset_function = FunctionOffset.create_with_args(offset) if offset is not None else None
        self.__limit_function = FunctionLimit.create_with_args(limit) if limit is not None else None
        self.__has_offset = False
        self.__has_limit = False

    # noinspection PyUnusedLocal
    def visit_function_offset(self, node, parents=None, context=None):
        self.__has_offset = True
        return self.__offset_function or node

    # noinspection PyUnusedLocal
    def visit_function_limit(self, node, parents=None, context=None):
        self.__has_limit = True
        return self.__limit_function or node

    def __call__(self, tree):
        result = self.visit(tree)
        if not self.__has_offset and self.__offset_function is not None:
            result = UnknownOperation(self.__offset_function, result)

        if not self.__has_limit and self.__limit_function is not None:
            result = UnknownOperation(self.__limit_function, result)

        return result


def get_split_vars(split):
    sql_vars = {}

    start = 0.
    for phase in ['train', 'test', 'validation']:
        percentage = split.get(phase)
        if phase is None:
            continue

        if percentage is None:
            sql_vars['phase_%s_start' % phase] = -1
            sql_vars['phase_%s_end' % phase] = -1
            continue

        sql_vars['phase_%s_start' % phase] = start
        sql_vars['phase_%s_end' % phase] = start + percentage

        start += percentage

    return sql_vars


class QueryParser(object):
    def __init__(self):
        self.__parser = parser()

    def parse_query(self, query):
        return self.__parser.parse(query, lexer=lexer())


def resolve_tree(tree, *transformers):
    transformers2 = [UnknownOperationResolver(OrOperation), MLQueryTransformer()]

    transformers2.extend(transformers)

    resolved_tree = None
    for transformer in transformers2:
        if resolved_tree is None:
            resolved_tree = transformer(tree)
            continue

        resolved_tree = transformer(resolved_tree)

    return resolved_tree


def visit_query(visitor, tree):
    resolver = resolve_tree(tree)
    visitor.visit(resolver)

    return visitor


def tree_to_sql_parts(tree, sql_helper):
    sql_builder = SQLQueryBuilder(sql_helper)
    visit_query(sql_builder, tree)

    return sql_builder.build_vars(), sql_builder.build_where()


if __name__ == '__main__':  # pragma: no cover
    QueryParser().parse_query('(type:Annotation OR type:Image) AND (bucket:1172 or bucket:3848 or bucket:2581 or bucket:269 or bucket:802)')



# -*- coding: utf8 -*-
from ...backend_mixin import BackendMixin
from .gcs_object_store import GCSObjectStore


class BackendGCSSignedUrlService(BackendMixin):
    def __init__(self, connection, config, session, handle_api):
        super(BackendGCSSignedUrlService, self).__init__(connection, config, session, handle_api)

    def get_signed_urls(self, methods, object_names, content_type=None, **kwargs):
        headers = []
        for key in sorted(kwargs.keys()):
            val = kwargs[key]
            headers.append('%s:%s' % (key, val))

        msg = {
            'methods': methods,
            'paths': object_names,
            'headers': headers,
        }

        if content_type:
            msg['content_type'] = content_type

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=self._volume_id)

        result = self._handle_api(self._config, self._session, 'post', url, msg)
        res = {method: result.get(method.lower(), []) for method in methods}

        return res


class BackendGCSObjectStore(BackendMixin, GCSObjectStore):
    def __init__(self, connection, config, session, handle_api):
        super(BackendGCSObjectStore, self).__init__(connection, config, session, handle_api)
        self.__bucket_name = None
        auth = connection.data_volume_config.object_store_config.get('auth')

        if auth != 'gcloud':
            self._signed_url_service = BackendGCSSignedUrlService(connection, config, session, handle_api)

    def __iter__(self):
        return super(BackendGCSObjectStore, self).__iter__()

    def close(self):
        super(BackendGCSObjectStore, self).close()

    @classmethod
    def _group_files_by_meta(cls, objects):
        content_type_grouped = {}
        for obj in objects:
            if obj.content_type not in content_type_grouped:
                content_type_grouped[obj.content_type] = []

            content_type_grouped[obj.content_type].append(obj)

        return content_type_grouped

    def _get_urls_for_paths(self, paths, content_type, headers):
        urls = self._signed_url_service.get_signed_urls(['HEAD', 'PUT'], paths, content_type, **headers)
        head_urls = urls['HEAD']
        put_urls = urls['PUT']
        return head_urls, put_urls

    def _upload_batch_async(self, content_type, files_info, callback):
        content_headers = self.get_content_headers()
        upload_paths = list(map(lambda x: GCSObjectStore._get_shafile_path(x.sha), files_info))
        if self._signed_url_service is not None:
            head_urls, put_urls = self._get_urls_for_paths(upload_paths, content_type, content_headers)
        else:
            put_urls = upload_paths
            head_urls = [None] * len(upload_paths)

        for cur_file, head_url, put_url in zip(files_info, head_urls, put_urls):
            self.upload_async(cur_file, head_url, put_url, callback=callback)

    def add_objects_async(self, objects, callback=None):
        grouped_files = self._group_files_by_meta(objects)
        for content_type in grouped_files:
            grouped_objects = grouped_files[content_type]
            self._upload_batch_async(content_type, grouped_objects, callback)

    @property
    def packs(self):
        return super(BackendGCSObjectStore, self).packs

    def contains_loose(self, sha):
        return super(BackendGCSObjectStore, self).contains_loose(sha)

    def contains_packed(self, sha):
        return super(BackendGCSObjectStore, self).contains_packed(sha)



import logging
from .base_aws import AwsBase
from .wrappers import handle_not_found

logger = logging.getLogger('aws.sts')


class Iam(AwsBase):
    def __init__(self, aws_ctx, org):
        super(Iam, self).__init__(aws_ctx, org)
        self.client = AwsBase.client('iam', aws_ctx)

    @handle_not_found
    def get_role(self, name):
        return self.client.get_role(RoleName=name)



from functools import wraps
from botocore import exceptions


def handle_not_found(fn):
    @wraps(fn)
    def try_error(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except exceptions.ClientError as e:
            # http://boto3.readthedocs.io/en/latest/reference/services/cloudformation.html#CloudFormation.Client.describe_stacks incorrectly indicates that `AmazonCloudFormationException` is supposed to be risen
            if e.args[0].endswith('does not exist') or '(NoSuchEntity)' in e.args[0]:
                return None

            raise

    return try_error



# -*- coding: utf8 -*-


def monitor_logs(ctx, url, disable_colors):
    from mali_commands.sse_firebase import LogsThread

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', url)

    logs_thread = LogsThread(ctx.obj.config, result['url'], disable_colors)
    logs_thread.start()
    logs_thread.join()



# -*- coding: utf8 -*-
import csv
import logging
import six

from ..api import default_api_retry
from .base_db_index import BaseMLIndex
from ..backend_mixin import BackendMixin


class BackendMLIndex(BackendMixin, BaseMLIndex):
    def __init__(self, connection, config, session, handle_api):
        self.__multi_process_control = None
        super(BackendMLIndex, self).__init__(connection, config, session, handle_api)

    def _create_table_if_needed(self):
        pass

    def set_entries_from_url(self, file_name, isolation_token):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage' % self._volume_id

            msg = {
                'index_url': file_name,
                'isolation_token': isolation_token,
            }

            self._handle_api(self._config, session, 'post', url, msg, retry=default_api_retry(), is_async=True)

    @staticmethod
    def _sync_set_entries_from_list(config, handle_api, session, volume_id, decoded_entries, isolation_token):
        rows = []
        for name, sha, ctime, mtime, mode, uid, gid, size, url in decoded_entries:
            row = {
                'name': name,
                'sha': sha,
                'ctime': ctime,
                'mtime': mtime,
                'mode': mode,
                'uid': uid,
                'gid': gid,
                'size': size,
                'url': url,
            }

            rows.append(row)

        url = 'data_volumes/%s/index/stage' % volume_id
        msg = {
            'entries': rows,
            'isolation_token': isolation_token,
        }

        handle_api(config, session, 'post', url, msg, retry=default_api_retry())

    def set_multi_process_control(self, multi_process_control):
        self.__multi_process_control = multi_process_control

    def set_entries_from_list_async(self, entries, isolation_token, callback=None):
        args = (self._config, self._handle_api, self._session, self._volume_id, list(self._decode_entries(entries)), isolation_token)

        self.__multi_process_control.execute(_sync_set_entries_from_list, args=args, callback=callback, high=True)

    def set_entries_from_list(self, entries, isolation_token):
        BackendMLIndex._sync_set_entries_from_list(self._config, self._handle_api, self._session, self._volume_id, self._decode_entries(entries), isolation_token)

    class ChangeSetIter(object):
        def __init__(self, data, data_type=None):
            self.__data = data
            data_type = data_type or 'csv'

            if data_type.lower() == 'csv':
                self.__reader = csv.DictReader(data)

        def __iter__(self):
            return self

        def next(self):
            data = next(self.__reader)
            return data['name'], data['op']

        def __next__(self):
            return self.next()

    def get_changeset(self, index_url, isolation_token):
        if not index_url:
            logging.debug('no data provided')
            return

        logging.debug('add data %s', index_url)

        def decode_iter(gen):
            for line in gen:
                if not isinstance(line, six.string_types):
                    line = line.decode()

                yield line

        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage' % self._volume_id

            msg = {
                'index_url': index_url,
                'dry_mode': True,
                'isolation_token': isolation_token,
            }

            result = self._handle_api(self._config, session, 'post', url, msg, retry=default_api_retry())

            r = session.get(result['change_set_url'], stream=True)
            r.raise_for_status()

            return self.ChangeSetIter(decode_iter(r.iter_lines()))

    def begin_commit(self, commit_sha, tree_id, ts):
        raise NotImplementedError(self.begin_commit)

    def end_commit(self):
        raise NotImplementedError(self.end_commit)


_sync_set_entries_from_list = BackendMLIndex._sync_set_entries_from_list



# -*- coding: utf8 -*-
try:
    import ConfigParser as configparser
except ImportError:
    import configparser

import json

import os
import requests
from requests.adapters import HTTPAdapter
import warnings
from .exceptions import NonRetryException


warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")


class GooglePackagesMissing(NonRetryException):
    pass


class GCPServices(object):
    CLOUD_SDK_CONFIG_DIR = 'CLOUDSDK_CONFIG'
    _CONFIG_DIRECTORY = 'gcloud'
    _CREDENTIALS_FILENAME = 'application_default_credentials.json'
    _WINDOWS_CONFIG_ROOT_ENV_VAR = 'APPDATA'
    _AUTHORIZED_USER_TYPE = 'authorized_user'
    _SERVICE_ACCOUNT_TYPE = 'service_account'
    _DEFAULT_CONFIG_FILE = 'config_default'
    _DEFAULT_CONFIG_PATH = 'configurations'
    _VALID_TYPES = (_AUTHORIZED_USER_TYPE, _SERVICE_ACCOUNT_TYPE)

    def __init__(self):
        pass

    @classmethod
    def _setup_session(cls, session):
        a = HTTPAdapter(pool_maxsize=50)
        b = HTTPAdapter(pool_maxsize=50)
        session.mount('http://', a)
        session.mount('https://', b)

    @classmethod
    def get_config_path(cls):
        # If the path is explicitly set, return that.
        try:
            return os.environ[cls.CLOUD_SDK_CONFIG_DIR]
        except KeyError:
            pass

        # Non-windows systems store this at ~/.config/gcloud
        if os.name != 'nt':
            return os.path.join(os.path.expanduser('~'), '.config', cls._CONFIG_DIRECTORY)
        # Windows systems store config at %APPDATA%\gcloud
        else:
            try:
                return os.path.join(os.environ[cls._WINDOWS_CONFIG_ROOT_ENV_VAR], cls._CONFIG_DIRECTORY)
            except KeyError:
                # This should never happen unless someone is really
                # messing with things, but we'll cover the case anyway.
                drive = os.environ.get('SystemDrive', 'C:')
                return os.path.join(drive, '\\', cls._CONFIG_DIRECTORY)

    @classmethod
    def get_application_default_credentials_path(cls):
        config_path = cls.get_config_path()
        return os.path.join(config_path, cls._CREDENTIALS_FILENAME)

    @classmethod
    def get_default_config_path(cls):
        config_path = cls.get_config_path()
        return os.path.join(config_path, cls._DEFAULT_CONFIG_PATH, cls._DEFAULT_CONFIG_FILE)

    @classmethod
    def get_default_project_id(cls):
        default_config = cls.get_default_config_path()
        config = configparser.ConfigParser()
        config.read(default_config)
        try:
            return config.get('core', 'project')
        except (configparser.NoOptionError, configparser.NoSectionError):
            return None

    @classmethod
    def _get_gcloud_auth_session(cls, credentials=None):
        try:
            import google.auth.transport.requests
        except ImportError:
            raise GooglePackagesMissing()

        credentials = credentials or cls.gcp_default_credentials()

        auth_method_session = google.auth.transport.requests.AuthorizedSession(credentials)
        cls._setup_session(auth_method_session)

        return auth_method_session, credentials

    @classmethod
    def gcp_default_credentials(cls):
        try:
            import google.auth.exceptions
            from google.oauth2 import service_account, credentials as google_credentials
        except ImportError:
            raise GooglePackagesMissing()

        filename = cls.get_application_default_credentials_path()
        try:
            with open(filename) as file_obj:
                info = json.load(file_obj)
                credential_type = info.get('type')
        except (IOError, OSError):
            return None

        if credential_type == cls._AUTHORIZED_USER_TYPE:
            return google_credentials.Credentials.from_authorized_user_info(info)
        elif credential_type == cls._SERVICE_ACCOUNT_TYPE:
            credentials = service_account.Credentials.from_service_account_info(info)
        else:
            raise google.auth.exceptions.DefaultCredentialsError(
                'The file {file} does not have a valid type. '
                'Type is {type}, expected one of {valid_types}.'.format(
                    file=filename, type=credential_type, valid_types=cls._VALID_TYPES))

        return credentials

    @classmethod
    def get_auth_session(cls, auth_method, credentials):
        if auth_method == 'gcloud':
            return cls._get_gcloud_auth_session(credentials)[0]

        thread_session = requests.Session()
        cls._setup_session(thread_session)

        return thread_session

    @classmethod
    def gcs_service(cls, credentials=None):
        try:
            from google.cloud import storage
        except ImportError:
            raise GooglePackagesMissing()

        auth_method_session, credentials = cls._get_gcloud_auth_session(credentials)

        params = {
            '_http': auth_method_session
        }

        if credentials is not None:
            params['credentials'] = credentials

        project = cls.get_default_project_id()
        if project is not None:
            params['project'] = project

        return storage.Client(**params)



import logging
from .base_aws import AwsBase
from .wrappers import handle_not_found

logger = logging.getLogger('CF')

notification_arn = 'arn:aws:sns:us-east-2:502812207255:t1'


class Cf(AwsBase):
    def __init__(self, aws_ctx, org):
        super(Cf, self).__init__(aws_ctx, org)
        self.client = AwsBase.client('cloudformation', aws_ctx)

    def create(self, stack_name, stack_data, parameters=None, tags=None):
        parameters = self.dict_to_tuple_array(parameters or {}, key='ParameterKey', value='ParameterValue', value_transformer=self.ensure_str)
        tags = self.dict_to_tuple_array(tags or {}, value_transformer=self.ensure_str)
        logger.info('Create: %s: %s', stack_name, stack_data)
        result = self.client.create_stack(
            StackName=stack_name,
            TemplateBody=stack_data,
            Parameters=parameters,
            Tags=tags,
            Capabilities=['CAPABILITY_NAMED_IAM'],
        )
        return result

    @handle_not_found
    def get(self, stack_name):
        response = self.client.describe_stacks(
            StackName=stack_name
        )
        stack = response['Stacks'][0]
        output_dict = self.tuple_array_do_dict(stack.get('Outputs', []), key='OutputKey', value='OutputValue')
        response = {'_src': stack, 'status': stack['StackStatus'], 'name': stack['StackName'], 'id': stack['StackId'], 'outputs': output_dict}
        return response

    @handle_not_found
    def resources(self, stack_name):
        response = self.client.describe_stack_resources(
            StackName=stack_name
        )
        stack = response['StackResources']
        output_dict = [{
            'LogicalResourceId': resource.get('LogicalResourceId'),
            'ResourceStatusReason': resource.get('ResourceStatusReason'),
            'ResourceStatus': resource.get('ResourceStatus')
        } for resource in stack]
        return output_dict



import logging
from .base_aws import AwsBase

logger = logging.getLogger('aws.sts')


class Sts(AwsBase):
    def __init__(self, aws_ctx, org):
        super(Sts, self).__init__(aws_ctx, org)
        self.client = AwsBase.client('sts', aws_ctx)

    def whoami(self):
        return self.client.get_caller_identity()



import boto3
import six


class AwsBase(object):
    def __init__(self, aws_ctx, org):
        self.ctx = aws_ctx
        self.org = org

    @classmethod
    def client(cls, name, aws_ctx):
        return boto3.client(name, region_name=aws_ctx.region)

    @classmethod
    def resource(cls, name, aws_ctx):
        return boto3.resource(name, region_name=aws_ctx.region)

    @classmethod
    def dict_to_tuple_array(cls, data, key='Key', value='Value', value_transformer=None):
        def default_transformation(x):
            return x

        if value_transformer is None:
            value_transformer = default_transformation
        return [{key: k, value: value_transformer(v)} for k, v in data.items()]

    @classmethod
    def tuple_array_do_dict(cls, data, key='Key', value='Value'):
        return {x[key]: x[value] for x in data}

    @classmethod
    def ensure_str(cls, value):
        if isinstance(value, bytes):
            return value.decode('utf-8')

        if isinstance(value, six.string_types):
            return value

        if value is None:
            return ''

        return str(value)



# -*- coding: utf8 -*-
import json
import logging
import os
import uuid
from traceback import format_exception_only
import six
import click
import yaml
from click import exceptions
import msgpack
from mali_commands.commons import output_result
from mali_commands.utils import monitor_logs
from .utilities import source_tracking
from mlcrypto import Asymmetric, MultiKeyEnvelope, SshIdentity
from .utilities.default_params import org_with_default


@click.group('run', help='runs an experiment on a cluster. By defaults run on a local cluster ')
def run_commands():
    pass


def _load_recipe(r_path):
    if not os.path.isfile(r_path):
        return {}
    else:
        print('loading defaults from recipe: %s' % r_path)
        with open(r_path) as f:
            return yaml.safe_load(f)


DEFAULT_RECIPE_PATH = '.ml_recipe.yaml'


def _source_tracking_data(path_=None):
    if path_ is None:
        path_ = os.getcwd()

    repo_obj = None
    src_data = {}
    try:
        repo_obj = source_tracking.get_repo(path_)
        src_data = _export_repo_state(repo_obj, mode='local')
    except Exception as ex:  # noinspection PyBroadException
        src_data['error'] = export_exception(ex)

    return src_data, repo_obj


def export_exception(ex):
    return ('\n'.join(format_exception_only(ex.__class__, ex))).strip()


def _export_repo_state(repo_obj, mode=None):
    mode = mode or 'local'
    src_data = {}
    try:
        src_data['branch'] = repo_obj.branch.name
        src_data['remote'] = repo_obj.remote_url
        src_data['sha_local'] = repo_obj.head_sha
        src_data['sha_local_url'] = repo_obj.head_sha_url
        src_data['is_dirty'] = not repo_obj.is_clean
        src_data['mode'] = mode

        commit_data = repo_obj.export_commit(repo_obj.repo.head.commit) if repo_obj.has_head else None

        if commit_data is not None:
            src_data.update(commit_data)

    # noinspection PyBroadException
    except Exception as ex:
        logging.exception('Failed to get repo state')
        src_data['error'] = export_exception(ex)

    logging.debug('_export_repo_state: %s', src_data)
    return src_data


def get_tracking_repo(repo_obj):
    repo_path = os.path.join(repo_obj.repo.working_dir, '.ml_tracking_repo')
    if os.path.isfile(repo_path):
        with open(repo_path) as f:
            tracking_repo = f.read().strip()
            logging.info('{} is tracked by {}'.format(repo_obj.repo.working_dir, tracking_repo))
            return tracking_repo
    # TODO: ADD QUERY to track server
    # return "git@github.com:missinglinkai/sim-test-remote-run.git"
    return None


def _validate_tracking_target(repo_obj, shadow_repo):
    if shadow_repo is None:
        return {'error': 'no tracking repository found.'}
    if shadow_repo.startswith('http'):
        return {'error': 'HTTP[S] git repositories are not supported. Please use the git/ssh version of {}'.format(shadow_repo)}
    return {'tracking_repository_url': shadow_repo}


def _sync_working_dir_if_needed(repo_obj, invocation_id):
    try:
        src_data = _validate_tracking_target(repo_obj, get_tracking_repo(repo_obj))
        if 'tracking_repository_url' in src_data:
            tracking_repository_url = src_data['tracking_repository_url']
            logging.info('Repository tracking is enabled. Tracking to repository: {}'.format(tracking_repository_url))
            source_tracking_repo = source_tracking.GitRepoSyncer.clone_tracking_repo(tracking_repository_url)
            commit_tag = source_tracking.GitRepoSyncer.merge_src_to_tracking_repository(repo_obj.repo, source_tracking_repo, br_tag=invocation_id)
            shadow_repo_obj = source_tracking.get_repo(repo=source_tracking_repo)
            cur_br = source_tracking_repo.active_branch
            source_tracking_repo.git.checkout(commit_tag)
            shadow_repo_obj.refresh()
            src_data = _export_repo_state(shadow_repo_obj, mode='shadow')
            source_tracking_repo.git.checkout(cur_br)

        if 'error' not in src_data:
            logging.info('Tracking repository sync completed. This experiment source code is available here: {}'.format(src_data['sha_local_url']))
        else:
            logging.info('Tracking repository sync Failed. The Error was: {}'.format(src_data['error']))

        return src_data

    except Exception as ex:
        ex_txt = export_exception(ex)
        logging.exception("Failed to init repository tracking. This experiment may not be tracked")
        return {'error': ex_txt}


def parse_env_array_to_dict(env_array):
    if env_array is None:
        return {}
    res = {}
    for env_tuple in env_array:
        key, value = env_tuple.split('=')
        key = key.strip()
        value = value.strip()

        if (value.startswith("'") and value.endswith("'")) or (value.startswith('"') and value.endswith('"')):
            value = value[1:-1]
        res[key] = value
    return res


def _build_run_args(kwargs):
    env = kwargs.pop('env', [])
    persistent_path = kwargs.pop('persistent_path', [])

    persistent_paths = None
    if len(persistent_path) > 0:
        persistent_paths = []
        for host_path, mount_path in persistent_path:
            persistent_paths.append(dict(host_path=host_path, mount_path=mount_path))

    input_data = {
        'org': kwargs.pop('org', None),
        'project': kwargs.pop('project', None),
        'image': kwargs.pop('image', 'missinglinkai/tensorflow:gce'),
        'git_repo': kwargs.pop('git_repo', None),
        'git_tag': kwargs.pop('git_tag', None),
        'source_dir': kwargs.pop('source_dir', None),
        'command': kwargs.pop('command', None),
        'data_query': kwargs.pop('data_query', None),
        'data_volume': kwargs.pop('data_volume', None),
        'data_use_iterator': kwargs.pop('data_iterator', False),
        'data_dest_folder': kwargs.pop('data_dest', None),
        'output_paths': kwargs.pop('output_paths', []),
        'persistent_paths': persistent_paths,
        'gpu': kwargs.pop('gpu', True),
        'git_identity': kwargs.pop('git_identity', None),
        'env': parse_env_array_to_dict(env)
    }
    input_data['env'] = json.dumps(input_data.get('env', {}))

    return input_data


def _read_input_data_and_load_recipe(kwargs):
    # Load Recipe
    recipe = kwargs.pop('recipe', None)
    recipe_data = _load_recipe(recipe or DEFAULT_RECIPE_PATH)
    for k, v in recipe_data.items():
        if k not in kwargs:
            kwargs[k] = v

    # Apply Defaults
    input_data = _build_run_args(kwargs)

    # Data test must reside inside /data as it will reside inside the data folder, until is configurable
    if not (input_data.get('data_dest_folder', ) or '/data').startswith('/data'):
        raise click.BadOptionUsage('`--data_dest` must begin with /data')

    if input_data.get('image') is None and input_data.get('command') is None:
        raise exceptions.BadOptionUsage('No command nor image provided')

    if input_data.get('project') is None:
        raise exceptions.BadOptionUsage('Please provide `project`')

    if input_data.get('org') is None:
        raise exceptions.BadOptionUsage('Please provide `org`')

    return input_data


def _has_valid_git_pointer(input_data):
    logging.debug('is valid git pointer? %s', input_data)
    is_valid = input_data.get('git_repo') is not None
    if is_valid:
        input_data['git_mode'] = input_data.get('git_mode', 'external')
    return is_valid


def _build_src_pointer(input_data):
    if not _has_valid_git_pointer(input_data):
        exp_id = uuid.uuid4().hex[:5]
        src_repo, x = _source_tracking_data(input_data.get('source_dir'))
        sync_res = _sync_working_dir_if_needed(x, exp_id)
        if 'error' in sync_res:
            logging.error(sync_res['error'])
        else:
            input_data['git_repo'] = sync_res['remote']
            input_data['git_tag'] = sync_res['branch']
            input_data['git_mode'] = sync_res['mode']

        if not _has_valid_git_pointer(input_data):  # still
            logging.warning('Failed to obtain git point. input was %s', input_data)
            raise exceptions.BadOptionUsage('Failed to obtain git point to use fo the experiment. please provide --git-repo, or --source-dir if you have shadow repository tracking enabled')
            # todo: find the current git path and get tracing path from the server if we don't have git point
            # todo better handling of ignoring

        return input_data


def _save_recipe(input_data, save_recipe):
    with open(save_recipe, 'w') as f:
        save_data = {}
        for k, v in input_data.items():
            if v is not None:
                save_data[k] = v
        yaml.safe_dump(save_data, f, default_flow_style=False)
        click.echo('Recipe saved to {}'.format(save_recipe))

    return True


@run_commands.command('xp')
@click.pass_context
@org_with_default()
@click.option('--project', type=int if six.PY3 else long, help='Project ID to hold the experiments started by the job')
@click.option('--image', type=str, required=False, help='Docker image to use, defaults to missinglinkai/tensorflow')
@click.option('--git-repo', type=str, required=False, help='Git repository to pull the code from')
@click.option('--git-tag', type=str, required=False, help='Git branch/tag for the git repository. defaults to master. The cloned code will be available under `/code`')
@click.option('--source-dir', type=str, required=False, help='source directory for the experiment (if you have configured tracking repository)')
@click.option('--command', type=str, required=False, help='command to execute')
@click.option('--gpu/--cpu', required=False, default=True, help='Deprecated.')
@click.option('--data-volume', type=str, required=False, help='data volume to clone data from')
@click.option('--data-query', type=str, required=False, help='query to execute on the data volume')
@click.option('--data-dest', type=str, required=False, help='destination folder and format for cloning data. If provided, must begin with /data')
@click.option('--data-iterator', type=bool, required=False, help='When set to True, data will not be cloned before the experiment and the quarry will be available for the SDK iterator')
@click.option('--recipe', '-r', type=click.Path(exists=True), required=False, help='recipe file. recipe file is yaml file with the `flag: value` that allows you to specify default values for all params for this function')
@click.option('--save-recipe', type=str, required=False, help='Saves a recipe for this call to the target file and quits. Note the default values are not encoded into the recipe')
@click.option('--env', multiple=True, default=None, required=False, help='Environment variables to pass for the invocation in key=value format. You can use this flag multiple times')
@click.option('--output-paths', multiple=True, required=False, help='Paths that will be exported to the Data management at the end of the invocation job. The paths will be available to the the running code under `/path_name` by defaults to `/output`')
@click.option('--git-identity', type=click.Path(exists=True), required=False, default=None,
              help='If provided, the provided path will be used as git (ssh) identity when pulling code. otherwise your default organisation identity will be used')
@click.option('--persistent-path', multiple=True, type=(str, str), required=False,
              help='` --persistent-path /mnt/data /my_data` \n'
                   'Maps a the `/mnt/data` path (or docker volume name) on the hosting server to the `/my_data` folder of the experiment. \n '
                   'Can be specified more than once.\n'
                   'Persistent mappings are given r/w permission.\n'
                   'USE WITH CAUTION!')
@click.option('--disable-colors', is_flag=True, default=None, required=False, help='Disable colors in logs')
@click.option('--attach', is_flag=True, default=None, required=False, help='Wait and print logs of the submitted job')
def run_experiment(ctx, **kwargs):
    kwargs = {k: v for k, v in kwargs.items() if v is not None and v != ()}
    save_recipe = kwargs.pop('save_recipe', False)
    disable_colors = kwargs.pop('disable_colors', False)
    attach = kwargs.pop('attach', False)

    input_data = _read_input_data_and_load_recipe(kwargs)
    _build_src_pointer(input_data)
    click.echo(err=True)
    click.echo('Job parameters:', err=True)
    for k, v in input_data.items():
        if v:
            click.echo('{:<20}: {}'.format(k, json.dumps(v)), err=True)

    click.echo(err=True)

    if save_recipe:
        _save_recipe(input_data, save_recipe)
        return

    org = input_data.pop('org')
    project = input_data.pop('project')

    _prepare_identity_file(ctx, org, input_data)

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'put', '{}/{}/invoke'.format(org, project), input_data)
    output_result(ctx, result, ['ok', 'invocation'])

    _attach(attach, ctx, org, result['invocation'], disable_colors)


def _attach(attach, ctx, org, job_id, disable_colors):
    if attach:
        url = '{org}/run/{job_id}/logs'.format(org=org, job_id=job_id)
        monitor_logs(ctx, url, disable_colors)


def _encode_dict(dict_):
    return msgpack.packb(dict_)


def _prepare_identity_file(ctx, org, input_data):
    if input_data['git_identity']:
        result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', '{}/run/security_credentials'.format(org))

        ssh_keys = [Asymmetric(key) for key in result['keys']]
        identity = SshIdentity(input_data['git_identity']).export_private_key_bytes()

        cipher = MultiKeyEnvelope(*ssh_keys)
        data = _encode_dict({
            'data': [
                {
                    'type': 'file',
                    'path': '@identity',
                    'data': identity
                }
            ],
            'encryption_path': result['keys']})

        # encrypt_string does not encode bytes
        input_data['encrypted_data'] = cipher.encrypt_string(data)


@run_commands.command('logs')
@org_with_default()
@click.option('--job-id', type=str)
@click.option('--disable-colors', is_flag=True, required=False)
@click.pass_context
def job_logs(ctx, org, job_id, disable_colors):
    url = '{org}/run/{job_id}/logs'.format(org=org, job_id=job_id)

    monitor_logs(ctx, url, disable_colors)



# -*- coding: utf8 -*-
import logging
import six
from requests import HTTPError
from requests.exceptions import MissingSchema, InvalidSchema
from .gcp_services import GCPServices, GooglePackagesMissing
from .exceptions import NonRetryException, NotFound, AccessDenied
from .path_utils import remove_moniker


class closing_with_condition(object):
    def __init__(self, thing, should_close):
        self.thing = thing
        self.should_close = should_close

    def __enter__(self):
        return self.thing

    def __exit__(self, *exc_info):
        if self.should_close:
            self.thing.close()


class GCSService(object):
    RETRYABLE_ERRORS = (IOError, )
    DEFAULT_MIMETYPE = 'application/octet-stream'
    NUM_RETRIES = 5

    def __init__(self, signed_url_service):
        self.signed_url_service = signed_url_service


class GCSDownload(GCSService):
    def __init__(self, auth_method, signed_url_service):
        super(GCSDownload, self).__init__(signed_url_service)
        self._auth_method = auth_method

    def download(self, object_name):
        import requests

        signed_urls = self.signed_url_service.get_signed_urls(['GET'], [object_name])
        url = signed_urls['GET'][0]

        r = requests.get(url)  # allowed to use requests
        r.raise_for_status()
        data = r.content

        logging.debug('downloaded  %s(%s)', object_name, len(data))

        return data


class GCSDownloadDirectDownload(GCSService):
    def __init__(self):
        super(GCSDownloadDirectDownload, self).__init__(None)

    @classmethod
    def download(cls, bucket_name, object_name):
        if bucket_name:
            bucket_name = remove_moniker(bucket_name)

            gcs = GCPServices.gcs_service()

            blob = gcs.bucket(bucket_name).blob(object_name)
            return blob.download_as_string()

        auth_session = GCPServices.get_auth_session('gcloud')

        r = auth_session.get(object_name)
        r.raise_for_status()

        return r.content


def _wrap_s3_call(callback, bucket_name, key):
    from botocore.exceptions import ClientError
    import boto3

    bucket_name = remove_moniker(bucket_name)

    s3_client = boto3.client('s3')

    try:
        return callback(s3_client, bucket_name, key)
    except s3_client.exceptions.NoSuchBucket:
        raise NotFound('No such S3 Bucket %s' % bucket_name)
    except ClientError as ex:
        if ex.response.get('Error', {}).get('Code') == 'RequestTimeout':
            raise

        if ex.response.get('Error', {}).get('Code') == 'AccessDenied':
            raise AccessDenied('Access Denied s3://%s/%s' % (bucket_name, key))

        if ex.response.get('Error', {}).get('Code') == 'NoSuchKey':
            raise NotFound('Object not found s3://%s/%s' % (bucket_name, key))

        raise NonRetryException('%s (s3://%s/%s)' % (ex, bucket_name, key))


def _open_file_with_exp(full_path_to_data):
    try:
        return open(full_path_to_data, 'rb')
    except (IOError, OSError):
        raise NonRetryException('failed to open %s' % full_path_to_data)


def _handle_file_object(full_path_to_data, callback_with_file_obj):
    if hasattr(full_path_to_data, 'read'):
        full_path_to_data.seek(0)
        should_close = False
        file_obj = full_path_to_data
    else:
        should_close = True
        file_obj = _open_file_with_exp(full_path_to_data)

    try:
        return callback_with_file_obj(file_obj)
    finally:
        if should_close:
            file_obj.close()


class S3DownloadDirectDownload(GCSService):
    def __init__(self):
        super(S3DownloadDirectDownload, self).__init__(None)

    @classmethod
    def download(cls, bucket_name, object_name):
        def s3_call(_s3_client, s3_bucket_name, s3_object_name):
            import boto3

            s3_resource = boto3.resource('s3')

            obj = s3_resource.Object(s3_bucket_name, s3_object_name)
            return obj.get()['Body'].read()

        return _wrap_s3_call(s3_call, bucket_name, object_name)


class GCSUploadDirect(GCSService):
    def __init__(self):
        super(GCSUploadDirect, self).__init__(None)

    @classmethod
    def upload(cls, credentials, bucket_name, object_name, full_path_to_data, headers):
        logging.info('gcs upload (direct) bucket: %s %s', bucket_name, object_name)

        try:
            import google.api_core.exceptions as google_exceptions
        except ImportError:
            raise GooglePackagesMissing()

        gcs = GCPServices.gcs_service(credentials)

        blob = gcs.bucket(bucket_name).blob(object_name)

        content_type = headers.get('Content-Type')

        try:
            _handle_file_object(full_path_to_data, lambda file_obj: blob.upload_from_file(file_obj, content_type))
        except google_exceptions.NotFound:
            raise NotFound('bucket %s not found' % bucket_name)
        except google_exceptions.PermissionDenied:
            raise AccessDenied('access denied to bucket %s' % bucket_name)


class S3UploadDirect(GCSService):
    def __init__(self):
        super(S3UploadDirect, self).__init__(None)

    # noinspection PyUnusedLocal
    @classmethod
    def upload(cls, bucket_name, object_name, full_path_to_data, headers, credentials=None):
        logging.info('s3 upload %s %s %s %s', bucket_name, object_name, full_path_to_data, headers)

        def s3_call(s3_client, s3_bucket_name, s3_object_name):
            _handle_file_object(full_path_to_data, lambda file_obj: s3_client.upload_fileobj(file_obj, s3_bucket_name, s3_object_name))

            logging.info('s3 uploaded %s %s', bucket_name, object_name)

        return _wrap_s3_call(s3_call, bucket_name, object_name)

    @classmethod
    def copy(cls, src, dest):
        logging.info('s3 copy %s => %s', src, dest)

        bucket_name, key = dest.split('/', 1)

        def s3_call(s3_client, s3_bucket_name, s3_key):
            s3_client.copy_object(Bucket=s3_bucket_name, CopySource=src, Key=s3_key)
            logging.info('s3 copied %s => %s', src, dest)

        logging.info('s3 copy %s => %s', src, dest)

        return _wrap_s3_call(s3_call, bucket_name, key)


class FileWithCallback(object):
    def __init__(self, file_obj, callback):
        file_obj.seek(0, 2)
        self._total = file_obj.tell()
        file_obj.seek(0)

        self._callback = callback
        self._file_obj = file_obj

    def __len__(self):
        return self._total

    def read(self, size):
        data = self._file_obj.read(size)
        if not six.PY2 and isinstance(data, six.string_types):
            data = data.encode()

        if six.PY3 and isinstance(data, six.string_types):
            data = data.encode()

        if self._callback is not None:
            self._callback(len(data))

        return data


class GCSUpload(GCSService):
    def __init__(self, credentials, auth_method, head_url, put_url):
        super(GCSUpload, self).__init__(None)
        self._head_url = head_url
        self._put_url = put_url
        self._auth_method = auth_method
        self._credentials = credentials

    def upload(self, full_path_to_data, headers, progress_callback=None):
        logging.info('gcs upload head: %s put: %s', self._head_url, self._put_url)

        auth_session = GCPServices.get_auth_session(self._auth_method, self._credentials)

        resp = None

        if self._head_url:
            logging.debug('try head url')
            resp = auth_session.head(self._head_url)

            logging.debug('try head resp %s', resp)

            if resp.status_code in (204, 404):
                logging.debug('file not found, uploading')
                resp = None

        if resp is None:
            def put_file(file_obj):
                file_obj_with_callback = FileWithCallback(file_obj, progress_callback)

                logging.debug('put url')
                put_resp = auth_session.put(self._put_url, data=file_obj_with_callback, headers=headers)
                logging.debug('put url done %s', put_resp)

                return put_resp

            resp = _handle_file_object(full_path_to_data, put_file)

        if resp.status_code in (401, 403):
            raise NonRetryException()

        resp.raise_for_status()


class GCSDeleteAll(GCSService):
    @classmethod
    def delete_all(cls, bucket_name, volume_id, max_files=None):
        try:
            import google.api_core.exceptions as google_exceptions
        except ImportError:
            raise GooglePackagesMissing()

        logging.info('delete all at %s/%s', bucket_name, volume_id)
        gcs = GCPServices.gcs_service()

        try:
            list_iter = gcs.bucket(bucket_name).list_blobs(prefix=volume_id)
        except google_exceptions.NotFound:
            logging.warning('bucket %s was not found', bucket_name)
            return

        total_deleted = 0
        for blob in list_iter:
            try:
                gcs.bucket(bucket_name).delete_blob(blob.name)
            except google_exceptions.NotFound:
                pass

            total_deleted += 1

            if max_files is not None and max_files == total_deleted:
                break

        logging .info('total deleted %s', total_deleted)

        return total_deleted


s3_moniker = 's3://'


def __retry_if_retry_possible_error(exception):
    logging.debug('got retry exception (upload/download) %s', exception)

    return not isinstance(exception, NonRetryException)


def _default_retry():
    from retrying import retry

    def decor(f):
        return retry(retry_on_exception=__retry_if_retry_possible_error, wait_exponential_multiplier=50, wait_exponential_max=5000)(f)

    return decor


class _CloudUploader(object):
    @classmethod
    @_default_retry()
    def with_retry_to_s3(cls, bucket_name, object_name, full_path_to_data, headers):
        bucket_name = remove_moniker(bucket_name)
        S3UploadDirect().upload(bucket_name, object_name, full_path_to_data, headers)

    @classmethod
    @_default_retry()
    def with_retry_transfer_s3(cls, bucket_name, object_name, full_path_to_data):
        full_s3_path = remove_moniker(full_path_to_data)
        object_name_with_bucket = remove_moniker(bucket_name) + '/' + object_name
        S3UploadDirect().copy(full_s3_path, object_name_with_bucket)

    @classmethod
    @_default_retry()
    def with_retry_to_gcs(cls, credentials, auth_method, put_url, head_url, headers, progress_callback, full_path_to_data):
        try:
            GCSUpload(credentials, auth_method, head_url, put_url).upload(full_path_to_data, headers, progress_callback)
        except (MissingSchema, InvalidSchema, MissingSchema):
            logging.exception('Invalid request')
            raise NonRetryException('Invalid request')
        except HTTPError as e:
            if e.response.status_code != 412:  # precondition
                return

            raise

    @classmethod
    @_default_retry()
    def with_retry_to_gcs_direct(cls, credentials, bucket_name, object_name, headers, full_path_to_data):
        try:
            GCSUploadDirect().upload(credentials, remove_moniker(bucket_name), object_name, full_path_to_data, headers)
        except HTTPError as e:
            if e.response.status_code != 412:  # precondition
                return

            raise


class _S3Upload(object):
    def __init__(self, full_path_to_data, headers):
        self._full_path_to_data = full_path_to_data
        self._headers = headers

    def upload(self, bucket_name, object_name):
        if self._full_path_to_data.startswith(s3_moniker):
            _CloudUploader.with_retry_transfer_s3(bucket_name, object_name, self._full_path_to_data)
            return

        _CloudUploader.with_retry_to_s3(bucket_name, object_name, self._full_path_to_data, self._headers)


class _GCSUpload(object):
    def __init__(self, full_path_to_data, headers):
        self._full_path_to_data = full_path_to_data
        self._headers = headers

    def upload(self, credentials, auth_method, bucket_name, object_name, head_url, put_url, callback):
        if put_url:
            _CloudUploader.with_retry_to_gcs(credentials, auth_method, put_url, head_url, self._headers, callback, self._full_path_to_data)
            return

        _CloudUploader.with_retry_to_gcs_direct(credentials, bucket_name, object_name, self._headers, self._full_path_to_data)


def do_upload(credentials, auth_method, bucket_name, object_name, full_path_to_data, headers, head_url, put_url, callback=None):
    if bucket_name is not None and bucket_name.startswith(s3_moniker):
        _S3Upload(full_path_to_data, headers).upload(bucket_name, object_name)
        return

    _GCSUpload(full_path_to_data, headers).upload(credentials, auth_method, bucket_name, object_name, head_url, put_url, callback)


def __handle_s3_download_if_needed(bucket_name, object_name):
    @_default_retry()
    def with_retry_s3():
        return S3DownloadDirectDownload().download(bucket_name, object_name)

    if object_name.startswith(s3_moniker):
        object_name = remove_moniker(object_name)
        bucket_name, object_name = object_name.split('/', 1)
        return with_retry_s3(), True
    elif bucket_name is not None and bucket_name.startswith(s3_moniker):
        return with_retry_s3(), True

    return None, False


def do_download(auth, bucket_name, object_name, signed_url_service=None):
    data, handled = __handle_s3_download_if_needed(bucket_name, object_name)

    if handled:
        return data

    @_default_retry()
    def with_retry():
        if signed_url_service:
            return GCSDownload(auth, signed_url_service).download(object_name)

        return GCSDownloadDirectDownload().download(bucket_name, object_name)

    return with_retry()


def do_delete_all(bucket_name, volume_id, max_files):
    return GCSDeleteAll(signed_url_service=None).delete_all(bucket_name, volume_id, max_files)



# -*- coding: utf8 -*-
import click
from .utilities import default_params
from terminaltables import PorcelainTable
from .commons import output_result


def dict_to_table(dict):
    keys = []
    values = []
    for k, v in dict.items():
        keys.append(k)
        values.append(v)

    return [{'key': k, 'value': v} for k, v in dict.items()]


@click.group('defaults')
def defaults_commands():
    pass


@defaults_commands.command('ls')
@click.pass_context
def defaults_commands_ls(ctx, **kwargs):
    output_result(ctx, dict_to_table(default_params.get_defaults()))


@defaults_commands.command('set')
@click.pass_context
@click.argument('key', type=str, required=True)
@click.argument('value', type=str, required=True)
def defaults_commands_set(ctx, **kwargs):
    default_params.set_default(kwargs['key'], kwargs['value'])
    output_result(ctx, dict_to_table(default_params.get_defaults()))


@defaults_commands.command('del')
@click.pass_context
@click.argument('key', type=str, required=True)
def defaults_commands_del(ctx, **kwargs):
    default_params.del_default(kwargs['key'])
    output_result(ctx, dict_to_table(default_params.get_defaults()))


@defaults_commands.command('get')
@click.pass_context
@click.argument('key', type=str, required=True)
def defaults_commands_del(ctx, **kwargs):
    val = default_params.get_default(kwargs['key'])
    output_result(ctx, {'key': kwargs['key'], 'value': val})



import click


def pop_key_or_prompt_if(kws, key, src_default=None, **kwargs):
    cur_val = kws.pop(key, src_default)
    while cur_val == src_default:
        if 'default' in kwargs and kws.get('silent', False):
            click.echo('{}: {} (silent)'.format(kwargs.get('text'), kwargs.get('default')))
            return kwargs['default']
        cur_val = click.prompt(**kwargs)
    return cur_val



# -*- coding: utf8 -*-
from mali_commands.resources import cloud_connector_defaults, get_ssh_path
from mali_commands.utilities.click_utils import pop_key_or_prompt_if
from six.moves.urllib import parse
from mali_commands.commons import output_result
import uuid
import logging
import click
from .sts import Sts
from .cloud_formation import Cf
from .iam import Iam
from click import exceptions
from .base_aws import AwsBase


class BackendContext(object):
    def __init__(self, ctx, kwargs):
        self.org = kwargs['org']
        self.ctx = ctx

    def resource_group_description(self, group_id):
        group_description = self._handle_api_call('get', '{}/aws/resource_group/{}'.format(self.org, group_id))

        return {k.pop('key'): k for k in group_description['data']}

    def put_resource_group_parameters(self, group_id, params, new_group=False):
        data = AwsBase.dict_to_tuple_array(params, key='key', value='values')
        group_description = self._handle_api_call('post' if new_group else 'put',
                                                  '{}/aws/resource_group/{}'.format(self.org, group_id),
                                                  data={'params': data})

        return {k.pop('key'): k for k in group_description['data']}

    def _handle_api_call(self, method, url, data=None):
        return self.ctx.obj.handle_api(self.ctx.obj, self.ctx.obj.session, method, url, data)


class AwsContext(BackendContext):
    def __init__(self, ctx, kwargs):
        BackendContext.__init__(self, ctx, kwargs)

        self.sts = Sts(ctx.obj.aws, self.org)
        self.whoami = self.sts.whoami()
        self.account_id = self.whoami['Account']
        self.region = self.ctx.obj.aws.region
        logging.debug('Org: %s, AWS account id: %s', self.org, self.account_id)
        self.auth_state = None
        self.refresh_auth_state()
        self.kwargs = kwargs
        self.cf = Cf(self.ctx.obj.aws, self.org)

    def refresh_auth_state(self):
        self.auth_state = self._handle_api_call('get', '{}/aws/authorisation_status/{}'.format(self.org, self.account_id))
        logging.info('AWS auth status', self.auth_state)
        if self.auth_state.get('auth_region') is not None and self.ctx.obj.aws.region is None:
            self.region = self.auth_state.get('auth_region')
            self.ctx.obj.aws.region = self.region
            logging.info('AWS region changed to %s', self.region)
        return self.auth_state

    def encrypt_and_send_connector(self):
        if self._authorised_and_configured():
            return

        self._prompt_for_ssh_if_needed()
        self._authorise_if_needed()
        kms_arn = self.auth_state.pop('encryption_key', None)
        if not kms_arn:
            raise click.exceptions.ClickException('You are authorised but have no KMS key configured. Please contact MissingLinkAi')

        self._init_prepare_connector_message(kms_arn, set_ssh=not self.auth_state['ssh_present'], set_mali=not self.auth_state['mali_config'])

        self.refresh_auth_state()

    def _authorised_and_configured(self):
        return self.auth_state['authorised'] and self.auth_state['ssh_present'] and self.auth_state['mali_config']

    def _prompt_for_ssh_if_needed(self):
        if self.auth_state['authorised'] and self.auth_state['ssh_present']:
            return

        self._pre_apply_default_ssh_key()
        has_key = False
        while not has_key:
            has_key = self._try_read_key()

    def _pre_apply_default_ssh_key(self):
        if self.kwargs.get('ssh_key_path') is None:
            self.kwargs['ssh_key_path'] = get_ssh_path()
            logging.info('Using %s as default ssh path', self.kwargs['ssh_key_path'])

    def _try_read_key(self):
        from mlcrypto import SshIdentity

        try:
            self.kwargs['ssh_key_path'] = pop_key_or_prompt_if(self.kwargs, 'ssh_key_path', text='SSH key path [--ssh-key-path]', default=get_ssh_path())
            ssh_key = SshIdentity(self.kwargs['ssh_key_path'])
            self.kwargs['ssh_private'] = ssh_key.export_private_key_bytes().decode('utf-8')
            self.kwargs['ssh_public'] = ssh_key.export_public_key_bytes().decode('utf-8')
            return True

        except Exception as e:
            click.echo('Failed to read ssh key from %s. Please check the path and try again. %s' % (self.kwargs['ssh_key_path'], str(e)))
            self.kwargs.pop('ssh_key_path')
        return False

    def _authorise_if_needed(self):
        if not self.auth_state['authorised']:
            self._ensure_role_absent()
            return self._get_submit_and_register_template('authorise_cf', 'ML-AUTH', parameters={'SshPublicKey': self.kwargs['ssh_public']})

    def _ensure_role_absent(self):
        iam = Iam(self.ctx.obj.aws, self.org)
        ml_role_name = 'MissingLinkResourceManager-{org}'.format(org=self.org)
        cur_role = iam.get_role(ml_role_name)
        if cur_role is not None:
            logging.info('_has_auth_stack: stack present %s', cur_role)
            raise exceptions.ClickException('Role %s already exists in your AWS account. Role ID: %s. Please contact MissingLinkAi' % (ml_role_name, cur_role['Role']['RoleId']))

    def _get_submit_and_register_template(self, cf_command, stack_name, parameters=None):
        stack_name = '{}-{}'.format(stack_name, self.org)
        self._ensure_stack_absent(stack_name)

        result = self._handle_api_call('get', '{}/aws/{}/{}'.format(self.org, cf_command, self.account_id))
        logging.debug(result)
        tags = AwsBase.tuple_array_do_dict(result['tags'], key='key', value='value')
        template = result['template']
        offer_id = result['offering_id']
        cf_response = self.cf.create(stack_name, stack_data=template, tags=tags, parameters=parameters)
        stack_id = cf_response['StackId']
        self._handle_api_call('put', '{}/aws/register_cf_stack/{}/{}?{}'.format(self.org, self.account_id, offer_id, parse.urlencode({'stack': stack_id})), {})
        self._wait_stack_for_complete(stack_name)
        self.refresh_auth_state()

    def _ensure_stack_absent(self, stack_name):
        cur_stack = self.cf.get(stack_name)
        if cur_stack is not None:
            logging.info('_has_stack_name: stack present %s', cur_stack)
            raise exceptions.ClickException('%s stack already exists in your AWS account with State: %s. Please contact MissingLinkAi' % (stack_name, cur_stack['status']))

    def _wait_stack_for_complete(self, stack):
        from time import sleep

        cur_stack = self.cf.get(stack)
        while not cur_stack['status'].lower().endswith('complete'):
            click.echo('%s: %s' % (stack, cur_stack['status']))
            resources = self.cf.resources(stack)
            output_result(self.ctx, resources)
            sleep(5)
            click.echo(chr(27) + "[2J")
            cur_stack = self.cf.get(stack)

        return self.cf.get(stack)

    @classmethod
    def get_kms(cls, kms_arn):
        from mlcrypto.legacy import KmsLegacyEnvelope

        return KmsLegacyEnvelope(kms_arn)

    @classmethod
    def encrypt(cls, kms, data):
        from mlcrypto.legacy import KmsLegacyEnvelope

        return KmsLegacyEnvelope.convert_encrypted_envelope_data_to_triple(kms.encrypt(data))

    @classmethod
    def decrypt(cls, kms, data):
        from mlcrypto.legacy import KmsLegacyEnvelope

        if len(data) == 3:
            iv, key, en_data = data
            return kms.decrypt(KmsLegacyEnvelope.convert_triple_to_encrypted_envelope_data({'iv': iv, 'key': key, 'data': en_data}))

    @classmethod
    def export_key_from_path(cls, ssh_key_path):
        from mlcrypto import SshIdentity

        return SshIdentity(ssh_key_path).export_private_key_bytes()

    def _init_prepare_connector_message(self, kms_key_arn, set_ssh, set_mali):
        template, config_data = cloud_connector_defaults(self.ctx, cloud_type='aws', kwargs=dict(connector=self.account_id))
        kms = self.get_kms(kms_key_arn)
        if set_ssh:
            ssh_key_path = pop_key_or_prompt_if(self.kwargs, 'ssh_key_path', text='SSH key path [--ssh-key-path]', default=get_ssh_path())
            ssh_key = self.export_key_from_path(ssh_key_path)
            ssh = self.encrypt(kms, ssh_key)
            template['ssh'] = ssh
        if set_mali:
            mali = self.encrypt(kms, config_data)
            template['mali'] = mali

        template['cloud_data'] = {}
        url = '{org}/cloud_connector/{name}'.format(org=self.org, name=template['name'])
        response = self._handle_api_call('post', url, template)
        print(response)

    def setup_vpc_if_needed(self):

        if self._is_vpc_active():
            return

        self._get_submit_and_register_template('new_vpc_cf', 'ML-VPC')

    def _is_vpc_active(self):
        if not self.region:
            return len(self.auth_state.get('vpc_regions', [])) > 0

        return self.region is not None and self.region in self.auth_state.get('vpc_regions', [])

    def setup_s3_if_needed(self, force=False):

        if not force and int(self.auth_state.get('s3_count') or 0) > 0:
            return

        return self._get_submit_and_register_template('new_s3_cf', 'ML-S3-{}'.format(uuid.uuid4().hex[:3].lower()))



# -*- coding: utf8 -*-
import click
from .commons import add_to_data_if_not_none, output_result


@click.group('orgs')
def orgs_commands():
    pass


@orgs_commands.command('list')
@click.pass_context
def list_orgs(ctx):
    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', 'orgs')

    output_result(ctx, result.get('orgs', []), ['org', 'display_name'])


@orgs_commands.command('create')
@click.option('--org', required=True)
@click.option('--displayName', required=False)
@click.option('--description', required=False)
@click.pass_context
def create_org(ctx, org, displayname, description):
    data = {}

    add_to_data_if_not_none(data, displayname, 'display_name')
    add_to_data_if_not_none(data, description, 'description')
    add_to_data_if_not_none(data, org, 'org')

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', 'orgs', data)

    output_result(ctx, result, ['ok'])


@orgs_commands.command('autoJoinDomain')
@click.option('--org', required=True)
@click.option('--domain', required=True, multiple=True)
@click.pass_context
def auto_join_domain(ctx, org, domain):
    data = {}

    add_to_data_if_not_none(data, list(domain), 'domains')

    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', 'orgs/{org}/autoJoin'.format(org=org), data)

    output_result(ctx, result, ['ok'])



# -*- coding: utf8 -*-
import logging
from collections import defaultdict
from functools import wraps
from pprint import pformat
import six
import botocore
import click
from click import exceptions
from colorama import Fore, Style

from .cloud.aws import AwsContext, BackendContext
from .commons import output_result
from .legit.context import Expando
from .utilities.default_params import org_with_default, option_with_default_factory


@click.group('aws')
@click.pass_context
@option_with_default_factory('--region', envvar="ML_AWS_REGION", help='AWS region to use', default_key='aws_region', required=False)
def aws_commands(ctx, **kwargs):
    ctx.obj.aws = Expando()
    ctx.obj.aws.region = kwargs.pop('region', None)


def handle_aws_errors(fn):
    @wraps(fn)
    def try_call(*args, **kwargs):
        try:
            return fn(*args, **kwargs)

        except botocore.exceptions.NoCredentialsError as ex:
            logging.info('Failed to validate authentication', exc_info=1)
            raise exceptions.ClickException('AWS configuration is incomplete. Please run `aws configure`. Error: %s' % str(ex))

    return try_call


@aws_commands.command('init')
@click.pass_context
@org_with_default()
@click.option('--ssh-key-path', default=None, help='SSH Key to use in for git clone commands, and connecting to virtual machines')
@handle_aws_errors
def init_auth(ctx, **kwargs):
    aws = AwsContext(ctx, kwargs)

    aws.encrypt_and_send_connector()
    aws.setup_vpc_if_needed()
    aws.setup_s3_if_needed()
    print(aws.auth_state)


def _trim_value(value, max_chars=30):
    if len(value) > max_chars:
        return value[:max_chars - 4] + ' ...'

    return value


def _with_format(text, fore='', style=''):
    return Style.RESET_ALL + fore + style + text + Style.RESET_ALL


class ArgRow:
    show_score = {
        'std': 1000,
        'adv': 500,
        'internal': 100
    }
    readonly_score = {
        False: 0,
        True: 50
    }
    configured_score = {
        False: 0,
        True: 5000
    }

    @classmethod
    def _values_to_str(cls, v):
        if isinstance(v, six.string_types):
            return v

        return '\n'.join(v if v is not None else '')

    def _active_value(self):
        return self.value if self.configured or self.read_only else self.default_value

    def __eq__(self, other):
        return self.name == other.name and self.arg == other.arg

    def __init__(self, name, arg):
        self.name = name
        self.arg = arg
        self.read_only = self.arg.get('read_only', False)
        self.configured = 'configured' in self.arg and not self.read_only
        self.show_level = self.arg.get('show_level', self.arg.get('edit_level', 'internal'))
        self.score = self.show_score[self.show_level] + self.configured_score[self.configured] + self.readonly_score[self.read_only]

        self.value = self._values_to_str(self.arg.get('configured'))
        self.default_value = self._values_to_str(self.arg.get('default'))
        self.active_value = self._active_value()

    color_legend = '  |  '.join(['Color Legend:', _with_format('read only', Fore.BLUE), _with_format('configured', Fore.GREEN), _with_format('default value', Fore.RESET)])
    param_filed = _with_format('Parameter', Fore.WHITE, Style.BRIGHT)
    value_field = _with_format('Value', Fore.WHITE, Style.BRIGHT)
    default_field = _with_format('Default', Fore.WHITE, Style.DIM)
    description_field = _with_format('Description', Fore.WHITE, Style.DIM)

    def configured_name(self):
        return _with_format(self.name, Fore.GREEN, Style.BRIGHT)

    def default_name(self):
        return _with_format(self.name, Fore.RESET, Style.BRIGHT)

    def row_name(self):
        return self.configured_name() if self.configured else self.default_name()

    def read_only_value(self):
        return _with_format(self.active_value, Fore.BLUE, Style.NORMAL)

    def configured_value(self):
        return _with_format(self.active_value, Fore.GREEN, Style.BRIGHT)

    def un_configured_value(self):
        return _with_format(self.active_value, Fore.RESET, Style.DIM)

    def configured_default_value(self):
        return _with_format(self.default_value, Fore.GREEN, Style.BRIGHT)

    def un_configured_default_value(self):
        return _with_format(self.default_value, Fore.RESET, Style.DIM)

    def row_value(self):
        if self.read_only:
            return self.read_only_value()

        if self.configured:
            return self.configured_value()

        return self.un_configured_value()

    def description_all(self):
        return _with_format(_trim_value(self.arg.get('description'), 60), Style.DIM)

    def row_description(self):
        return self.description_all()

    def row_default(self):
        if self.read_only:
            return ''

        if self.configured:
            return self.un_configured_default_value()

        return self.configured_default_value()

    def to_row(self):
        return {
            self.param_filed: self.row_name(),
            self.value_field: self.row_value(),
            self.default_field: self.row_default(),
            self.description_field: self.row_description(),
        }

    def displayed(self, target_show_levels, configured_only):
        if self.configured:
            return True

        shown = self.show_level in target_show_levels
        return shown and not configured_only

    @classmethod
    def import_and_filter(cls, data, target_show_levels, configured_only):
        res = []
        for k, v in data.items():
            ob = cls(k, v)
            if ob.displayed(target_show_levels, configured_only):
                res.append(ob)

        return res

    @classmethod
    def table_fields(cls, kwargs):
        show_defaults = kwargs.get('show_defaults', False)
        show_description = kwargs.get('show_description', False)

        fields = [cls.param_filed, cls.value_field]
        if show_defaults:
            fields.append(cls.default_field)
        if show_description:
            fields.append(cls.description_field)

        return fields

    @classmethod
    def print_table(cls, ctx, kwargs, data):
        table = [x.to_row() for x in sorted(data, key=lambda x: x.score, reverse=True)]
        fields = cls.table_fields(kwargs)
        click.echo(cls.color_legend)
        click.echo()
        output_result(ctx, table, fields=fields)

    @classmethod
    def parse_user_input(cls, tuples):
        res = defaultdict(list)
        for tup in tuples:
            res[tup[0]].append(tup[1])
            if 'None' in res[tup[0]]:
                res[tup[0]] = ['None']
        return dict(res)


@aws_commands.command('resource_group')
@click.pass_context
@org_with_default()
@click.argument('group', required=True, nargs=1)
@click.option('--json', required=False, default=False, help="output data in json format")
@click.option('--advanced', required=False, default=False, help="show advanced configuration parameters as well")
@click.option('--show-defaults/--no-defaults', required=False, default=False, help="also show default values")
@click.option('--create/--update', required=False, default=False, help="create new group or update existing")
@click.option('--show-description/--no-description', required=False, default=False, help="also show parameter descriptions")
@click.option('--configured-only/--configured-and-defaults', required=False, default=False, help="show only parameters with configured values")
@click.option('--set', required=False, default=None, type=(str, str), multiple=True, help="Change parameter value. Some parameters can be specified multiple times for arrays and you can remove configured option by setting it to `None`")
def aws_resource_group(ctx, **kwargs):
    group_id = kwargs.pop('group')

    json = kwargs.get('json', False)

    configured_only = kwargs.get('configured_only', False)
    target_show_levels = ['std']

    if kwargs['advanced']:
        target_show_levels = ['std', 'adv']
    aws = BackendContext(ctx, kwargs)

    new_values = ArgRow.parse_user_input(kwargs['set'])
    if new_values:
        response = aws.put_resource_group_parameters(group_id, new_values, new_group=kwargs.pop('create'))
    else:
        response = aws.resource_group_description(group_id)
    data = ArgRow.import_and_filter(response, target_show_levels, configured_only)
    if json:
        click.echo(pformat({v.name: v.arg for v in data}))
    else:
        ArgRow.print_table(ctx, kwargs, data)



# -*- coding: utf8 -*-
import logging
from collections import OrderedDict

from ...gcp_services import GCPServices
from ...gcs_utils import do_upload, do_delete_all, do_download, s3_moniker
from ...connection_mixin import ConnectionMixin
from ...dulwich.object_store import BaseObjectStore
from ...dulwich.objects import hex_to_filename, Blob


class GCSObjectStore(ConnectionMixin, BaseObjectStore):
    def __init__(self, connection):
        super(GCSObjectStore, self).__init__(connection)
        self.__upload_pool = None
        self._object_store_auth = connection.data_volume_config.object_store_config.get('auth')
        self.__bucket_name = connection.data_volume_config.object_store_config.get('bucket_name')
        self.__embedded = connection.data_volume_config.embedded
        self.__volume_id = self._connection.data_volume_config.volume_id
        self._signed_url_service = None
        self.__multi_process_control = None
        self.__gcp_credentials = None

    def delete_all(self, max_files=None):
        return do_delete_all(self.__bucket_name, self.__volume_id, max_files)

    def set_multi_process_control(self, multi_process_control):
        self.__multi_process_control = multi_process_control

    @classmethod
    def get_content_headers(cls, content_type=None, is_s3=False):

        headers = OrderedDict()
        if content_type:
            headers['Content-Type'] = content_type

        if not is_s3:
            headers['x-goog-acl'] = 'public-read'
            headers['x-goog-if-generation-match'] = '0'

        return headers

    @classmethod
    def _get_shafile_path(cls, sha):
        # Check from object dir
        return hex_to_filename('objects', sha)

    @classmethod
    def on_upload_error(cls, ex):
        raise ex

    def upload_async(self, obj, head_url=None, put_url=None, callback=None):
        args = self._gen_upload_sync_args(obj, head_url, put_url)

        def on_finish(result):
            callback(obj)

        self.__multi_process_control.execute(do_upload, args=args, callback=on_finish if callback else None)

    def _load_gcp_credentials(self):
        if self.__gcp_credentials is None:
            self.__gcp_credentials = GCPServices.gcp_default_credentials()

        return self.__gcp_credentials

    @property
    def _is_s3_bucket(self):
        return self.__bucket_name and self.__bucket_name.startswith(s3_moniker)

    def _gen_upload_sync_args(self, obj, head_url=None, put_url=None):
        object_name = self._get_shafile_path(obj.sha)

        if self.__bucket_name and not self._is_s3_bucket or self._object_store_auth == 'gcloud':
            credentials = self._load_gcp_credentials()
        else:
            credentials = None

        content_type = obj.content_type
        headers = GCSObjectStore.get_content_headers(content_type, self._is_s3_bucket)

        object_name = '%s/%s' % (self.__volume_id, object_name)

        return credentials, self._object_store_auth, self.__bucket_name, object_name, obj.full_path, headers, head_url, put_url,

    def _get_loose_object(self, metadata):
        logging.debug('get object %s', metadata)

        sha = metadata['@id']

        if self.__embedded or self.__bucket_name is not None:
            object_name = '%s/%s' % (self.__volume_id, self._get_shafile_path(sha))
        else:
            object_name = metadata['@url']

        data = do_download(
            self._object_store_auth, self.__bucket_name, object_name, signed_url_service=self._signed_url_service)

        blob = Blob()
        blob.set_raw_chunks([data], sha)
        return blob

    def get_raw(self, metadata):
        """Obtain the raw text for an object.

        :param metadata: metadata for the object.
        :return: tuple with numeric type and object contents.
        """
        ret = self._get_loose_object(metadata)
        if ret is not None:
            return ret.type_num, ret.as_raw_string()

        raise KeyError(metadata)

    @property
    def packs(self):
        raise NotImplementedError(self.packs)

    def __iter__(self):
        raise NotImplementedError(self.__iter__)

    def add_objects_async(self, objects, callback=None):
        for obj in objects:
            self.upload_async(obj, callback=callback)

    def contains_packed(self, sha):
        raise NotImplementedError(self.contains_packed)

    def contains_loose(self, sha):
        raise NotImplementedError(self.contains_loose)



import click
import os
import json
import logging
from .os_utils import mkdir_recursive


def _config_folder_path():
    return os.path.join(os.path.expanduser('~'), '.MissingLinkAI')


def _defaults_path():
    return os.path.join(_config_folder_path(), 'defaults.json')


def get_defaults():
    if os.path.isfile(_defaults_path()):
        try:
            with open(_defaults_path(), 'r') as f:
                return json.load(f)
        except Exception:
            logging.warning('Defaults file at %s is corrupted', _defaults_path(), exc_info=1)
    return {}


def _set_defaults(data):
    mkdir_recursive(_config_folder_path())
    with open(_defaults_path(), 'w') as f:
        json.dump(data, f)


def set_default(key, value):
    new_doc = get_defaults()
    new_doc[str(key)] = str(value)
    try:
        _set_defaults(new_doc)
    except Exception:
        logging.exception('Failed to save defaults to %s', _defaults_path())


def del_default(key):
    new_doc = get_defaults()
    val = new_doc.pop(key, None)
    if val is not None:
        logging.debug('Removing %s default value %s', key, val)
    try:
        _set_defaults(new_doc)
    except Exception:
        logging.exception('Failed to save defaults to %s', _defaults_path())


def get_default(val):
    return get_defaults().get(val)


def option_with_default_factory(param_name, **kwargs):
    default_key = kwargs.pop('default_key', None)
    if default_key is not None:
        kwargs['default'] = get_default(default_key)
    return click.option(param_name, **kwargs)


def org_with_default(**kwargs):
    defaults = dict(type=str, envvar='ML_ORG', default_key='org', required=True, help='Organisation')
    defaults.update(kwargs)
    return option_with_default_factory('--org', **defaults)


def project_with_default(**kwargs):
    defaults = dict(type=str, envvar='ML_PROJECT', default_key='project', required=True, help='Project')
    defaults.update(kwargs)
    return option_with_default_factory('--project', **defaults)



# -*- coding: utf8 -*-
import datetime
import json
import logging
import socket
import threading
import abc
import time
import requests
from colorama import Fore, Style
from sseclient import SSEClient
from requests import HTTPError


class ClosableSSEClient(SSEClient):
    """
    Hack in some closing functionality on top of the SSEClient
    """

    def __init__(self, url, session, *args, **kwargs):
        self.should_connect = True
        super(ClosableSSEClient, self).__init__(url, session=session, *args, **kwargs)

    def _connect(self):
        if self.should_connect:
            super(ClosableSSEClient, self)._connect()
        else:
            raise StopIteration()

    # noinspection PyProtectedMember
    def close(self):
        self.should_connect = False
        self.retry = 0
        # dig through the sseclient library to the requests library down to the underlying socket.
        # then close that to raise an exception to get out of streaming. I should probably file an issue w/ the
        # requests library to make this easier
        self.resp.raw._fp.fp._sock.shutdown(socket.SHUT_RDWR)
        self.resp.raw._fp.fp._sock.close()


class SSEThread(threading.Thread):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        self._sse = None
        self._sleeping_since = None
        super(SSEThread, self).__init__(name=name)

    @abc.abstractmethod
    def create_sse_client(self):
        pass

    @abc.abstractmethod
    def on_sse_event(self, path, data):
        pass

    def _handle_data(self, data):
        if data is None:
            return

        msg_data = json.loads(data)

        if msg_data is None:
            return

        self.on_sse_event(**msg_data)

    def _sse_loop(self):
        for msg in self._sse:
            if self._sleeping_since is not None:
                logging.info("server back online, total downtime %s", datetime.datetime.utcnow())
                self._sleeping_since = None

            self._handle_data(msg.data)

    def _inner_run_loop(self):
        try:
            self._sse = self.create_sse_client()
            self._sse_loop()
        except (EOFError, HTTPError):
            if self._sleeping_since is not None:
                self._sleeping_since = datetime.datetime.utcnow()

            logging.exception("error during processing, sleep and retry")
            time.sleep(1)
        except Exception:
            logging.exception("error during processing crashing")
            raise

    def run(self):
        while True:
            self._inner_run_loop()

    def close(self):
        if self._sse is not None:
            self._sse.close()


class LogsThread(SSEThread):
    def __init__(self, config, logs_endpoint, disable_colors):
        self._url = logs_endpoint
        self._session = requests.session()
        self._disable_colors = disable_colors
        self._config = config
        super(LogsThread, self).__init__(name='logs')

    def create_sse_client(self):
        return ClosableSSEClient(self._url, self._session)

    def __get_color(self, level):
        if self._disable_colors:
            return ''

        color = Style.RESET_ALL + Fore.WHITE

        levels = {
            'ERROR': Style.RESET_ALL + Fore.RED,
            'INFO': Style.RESET_ALL + Style.DIM,
            'WARNING': Style.RESET_ALL + Fore.YELLOW
        }

        return levels.get(level, color)

    @classmethod
    def __get_message(cls, category, level):
        if category is not None and level is not None:
            msg = '{ts} [{category} {level}] {message}'
        else:
            msg = '{ts} {message}'

        return msg

    def on_sse_event(self, path, data):
        if data is None:
            return

        if path == '/':  # snapshot
            values = sorted(data.values(), key=lambda d: d['ts'])
        else:
            values = [data]

        for val in values:
            msg = self.__get_message(val.get('category'), val.get('level'))

            color = self.__get_color(val.get('level'))
            print(color + msg.format(**val))



# -*- coding: utf8 -*-
import copy
import logging
import multiprocessing
import threading
import uuid
from multiprocessing import Pool
import os
import signal


DEFAULT_USE_THREADS = True


def process_worker_init(parent_id):
    def sig_int(_signal_num, _frame):
        os.kill(parent_id, signal.SIGSTOP)

    def sig_int_sef_fault(_signal_num, _frame):
        os.kill(parent_id, signal.SIGTERM)

    signal.signal(signal.SIGINT, sig_int)
    signal.signal(signal.SIGSEGV, sig_int_sef_fault)


class _MultiProcessControlShim(object):
    class MultiShimFuture(object):
        def wait(self):
            pass

    @classmethod
    def execute(cls, method, args, callback=None, high=False):
        method(*args)
        if callback is not None:
            callback(None)

        return cls.MultiShimFuture()

    def terminate(self):
        pass

    def close(self):
        pass

    def join(self):
        pass


def get_multi_process_control(processes, use_threads=None):
    if processes == 1:
        return _MultiProcessControlShim()

    if use_threads is None:
        use_threads = DEFAULT_USE_THREADS

    if use_threads is True:
        use_processes = os.environ.get('MALI_PROCESSES', '0') == '1'
        use_threads = not use_processes

    return _MultiProcessControl(processes, use_threads)


class _MultiProcessControl(object):
    def __use_threads(self, processes):
        from multiprocessing.pool import ThreadPool

        self.__using_threads = True
        self.__processing_pool = ThreadPool(processes)
        self.__processing_pool_high = ThreadPool(1)

    def __init__(self, processes, use_threads):
        processes = multiprocessing.cpu_count() * 5 if processes == -1 else processes
        self.__using_threads = False
        self.__controllers = multiprocessing.Semaphore(processes)
        self.__pending_jobs = []
        self.__processing_pool_high = None

        if use_threads:
            self.__use_threads(processes)
        else:
            try:
                self.__processing_pool = Pool(processes, process_worker_init, initargs=(os.getpid(),), maxtasksperchild=1)
            except AssertionError:
                self.__use_threads(processes)

        self._jobs_lock = threading.Lock()
        self._jobs = {}

    def join(self):
        logging.debug('%s pool joining', self.__class__)
        self.__wait_pending_jobs()
        logging.debug('%s pool joined', self.__class__)

    def close(self):
        if self.__processing_pool is not None:
            logging.debug('%s closing & joining pool', self.__class__)
            self.__processing_pool.close()
            self.join()

    def __wait_pending_jobs(self):
        while True:
            if len(self._jobs) == 0:
                break

            with self._jobs_lock:
                jobs = copy.copy(self._jobs)

            for token, async_result in jobs.items():
                if async_result is not None:
                    async_result.wait()

            self.__check_pending_jobs()

    def __check_pending_jobs(self):
        updates_jobs = False

        with self._jobs_lock:
            jobs = copy.copy(self._jobs)

        for token, async_result in jobs.items():
            if async_result is None or not async_result.ready():
                continue

            async_result.get()

            self._jobs[token] = None
            updates_jobs = True

        if updates_jobs:
            with self._jobs_lock:
                self._jobs = {token: async_result for token, async_result in self._jobs.items() if async_result is not None}

    def execute(self, method, args, callback=None, high=False):
        self.__check_pending_jobs()

        token = uuid.uuid4()
        processing_pool = self.__processing_pool_high if high and self.__processing_pool_high else self.__processing_pool

        job_async_result = processing_pool.apply_async(
            method,
            args=args,
            callback=callback)

        self._jobs[token] = job_async_result

        return job_async_result

    def terminate(self):
        self.__processing_pool.terminate()
        self._jobs = {}



# -*- coding: utf8 -*-


def get_keys(d):
    if isinstance(d, dict):
        return sorted(d.keys())

    if not d:
        return []

    return sorted(d[0].keys())


def dict_to_csv(d, fields):
    def get_row():
        for key in total_keys:
            yield obj.get(key, '')

    if isinstance(d, dict):
        d = [d]

    total_keys = fields

    items_iter = iter(d)

    obj = next(items_iter)
    if total_keys is None:
        total_keys = get_keys([obj])

    yield total_keys

    while True:
        yield list(get_row())
        try:
            obj = next(items_iter)
        except StopIteration:
            break


def format_json_data(json_data, formatters=None):
    if formatters is None:
        return json_data

    for row in json_data:
        for field, formatter in formatters.items():
            if field in row:
                row[field] = formatter(row[field])

    return json_data



# -*- coding: utf8 -*-
import logging
import click
import six

from .commons import output_result
from .utilities.click_utils import pop_key_or_prompt_if


def get_home_path():
    from os.path import expanduser
    return expanduser("~")


def get_ssh_path():
    from os.path import join
    return join(get_home_path(), '.ssh', 'id_rsa')


DOCKER_IMAGE = 'docker:latest'


def pull_image_from_image(client, image):
    import docker.errors as docker_errors

    try:
        client.images.get(DOCKER_IMAGE)
    except docker_errors.NotFound:
        click.echo('Pulling docker image')
        client.images.pull(DOCKER_IMAGE)

    cmd = 'docker pull {}'.format(image)
    socket_volumes = {'/var/run/docker.sock': {'bind': '/var/run/docker.sock'}}

    cont = client.containers.run(DOCKER_IMAGE, command=cmd, auto_remove=True, volumes=socket_volumes, environment={'ML_RM_MANAGER': '1'}, detach=True)
    logger_handler = cont.logs(stdout=True, stderr=True, stream=True)
    for log in logger_handler:
        logging.info(log)

    return client.images.get(image)


def auth_resource(ctx, org):
    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', '{org}/resource/authorise'.format(org=org))
    return result.get('token')


def validate_and_get_docker_client():
    import docker.errors as docker_errors
    import docker
    import requests

    client = docker.from_env()
    try:
        client.ping()
    except docker_errors.DockerException as ex:
        raise click.BadArgumentUsage('Failed to connect to docker host %s' % (str(ex)))
    except requests.exceptions.ConnectionError as ex:
        raise click.BadArgumentUsage('Failed to connect to docker host %s' % (str(ex)))

    logging.info('Docker host verified')

    return client


def pull_ml_image(config, docker_client):
    click.echo('Getting/updating MissingLinks Resource Manager image')
    img = pull_image_from_image(docker_client, config.rm_manager_image)
    return img


def _get_combined_volume_path(*args):
    res = {}
    for a in args:
        res.update(a)

    return res


def _docker_present_return_instance(command, *args, **kwargs):
    import docker.errors as docker_errors
    try:
        return command(*args, **kwargs)
    except docker_errors.NotFound:
        pass


def _docker_present(command, *args, **kwargs):
    import docker.errors as docker_errors
    try:
        return command(*args, **kwargs) or True  # we are looking only for exceptions here
    except docker_errors.NotFound:
        return False


def validate_running_resource_manager(config, docker_client, force):
    current_rm = _docker_present_return_instance(docker_client.containers.get, config.rm_container_name)
    if current_rm is None:
        return

    if not force:
        raise click.BadOptionUsage('Can not install resource manger while one is running. run `docker kill {}` do stop and reuse config or re-run with `--force` flag to clear all configuration'.format(current_rm.name))

    click.echo('Killing current Resource Manger (%s) due to --force flag' % current_rm.id)
    if current_rm.status == 'running':
        current_rm.kill()

    current_rm.remove(force=True)


def get_config_prefix_and_file(config):
    with open(config.config.config_file_abs_path, 'rb') as f:
        config_data = f.read()
    prefix = None
    if config.config.config_prefix is not None:
        prefix = config.config.config_prefix
    return prefix, config_data


ADMIN_VOLUME = {'/var/run/docker.sock': {'bind': '/var/run/docker.sock'}}


def _apply_config_to_volume(config, docker_client, ssh_key, token, prefix=None, config_data=None, ml_backend=None):
    if token is None and ssh_key is None:
        return

    config_volume = {config.rm_config_volume: {'bind': '/config'}}
    conf_mounts = _get_combined_volume_path(ADMIN_VOLUME, config_volume)

    ws_server = config.rm_socket_server

    if prefix is None and config_data is None:
        id_token = config.id_token
        if id_token is None:
            # TODO: make backend data commands support resource token
            raise click.BadOptionUsage('Please call mali auth init first')
        prefix, config_data = get_config_prefix_and_file(config)

    if config_data and six.PY3 and isinstance(config_data, bytes):
        config_data = config_data.decode('utf-8')

    command = ['config', '--ml-server', ws_server, '--ml-config-file', config_data]

    if prefix is not None:
        command.extend(['--ml-config-prefix', prefix])

    if token is not None:
        command.append('--ml-token')
        command.append(token)

    if ml_backend is not None:
        command.append('--ml-backend')
        command.append(ml_backend)

    if ssh_key is not None:
        command.append('--ssh-private-key')
        command.append(ssh_key)

    cont = docker_client.containers.run(config.rm_manager_image, command=command, volumes=conf_mounts, environment={'ML_RM_MANAGER': '1'}, detach=True)
    exit_code = cont.wait()
    if exit_code != 0:
        click.echo(cont.logs())
    cont.remove()


def _handle_token_and_data_path(ctx, force, org, token=None):
    cur_config = ctx.obj.config.resource_manager_config
    if force:
        click.echo('Current host config is deleted due to `--force` flag')
        cur_config = {}

    new_token = token or cur_config.get('token')

    if new_token is None:
        new_token = auth_resource(ctx, org)
    ctx.obj.config.update_and_save({
        'resource_manager': {
            'token': new_token,
        }
    })
    return new_token


def _validate_apply_config(ctx, docker_client, org, force, ssh_key_path, token):
    config = ctx.obj
    config_volume_name = config.rm_config_volume

    if force and _docker_present(docker_client.volumes.get, config_volume_name):
        click.echo('Deleting config volume (%s) due to --force flag')
        docker_client.volumes.get(config_volume_name).remove(force=True)

    new_image = not _docker_present(docker_client.volumes.get, config_volume_name)

    ssh_key = None
    if new_image:
        docker_client.volumes.create(config_volume_name)
        if ssh_key_path is None:
            ssh_key_path = click.prompt(text='SSH key path (--ssh-key-path)', default=get_ssh_path())

    token = _handle_token_and_data_path(ctx, force, org, token=token)

    if ssh_key_path is not None:
        from .cloud.aws import AwsContext
        ssh_key = AwsContext.export_key_from_path(ssh_key_path)

    _apply_config_to_volume(config, docker_client, ssh_key, token)
    return


def _validate_config_volume(config, docker_client):
    if not _docker_present(docker_client.volumes.get, config.rm_config_volume):
        raise click.BadArgumentUsage('Configuration volume is missing. Please re-install')


def _run_resource_manager(config, docker_client):
    _validate_config_volume(config, docker_client)
    click.echo('Starting Resource Manager')
    config_volume = {config.rm_config_volume: {'bind': '/config'}}
    run_mounts = _get_combined_volume_path(ADMIN_VOLUME, config_volume)
    return docker_client.containers.run(
        config.rm_manager_image,
        command=['run'],
        auto_remove=False,
        restart_policy={"Name": 'always'},
        volumes=run_mounts,
        environment={'ML_RM_MANAGER': '1', 'ML_CONFIG_VOLUME': config.rm_config_volume},
        detach=True,
        network='host',
        name=config.rm_container_name)


def get_selected_fields(client_input, default_fields=None):
    display_fields = default_fields
    if client_input is not None:
        selected_fields = [x.strip() for x in client_input.split(',')]
        if len(selected_fields) == 1 and selected_fields[0] == '*':
            display_fields = None
        else:
            display_fields = selected_fields
    return display_fields


def append_optional_filters_to_url(url, **kwargs):
    filters = ['{}={}'.format(k, v) for k, v in kwargs.items() if v is not None]
    if len(filters) > 0:
        url = '{url}?{filters}'.format(url=url, filters='&'.join(filters))
    return url


@click.group('resources', help='Resource Management')
def resource_commands():
    pass


@resource_commands.command('install')
@click.pass_context
@click.option('--org', type=str, help='organization to use', required=True)
@click.option('--ssh-key-path', type=str, help='Path to the private ssh key to be used by the resource manager', default=None)
@click.option('--force/--no-force', default=False, help='Force installation (stops current resource manager if present')
@click.option('--resource-token', default=None, type=str, help='MissingLink resource token. One will be generated if this instance of MALI is authorized')
def install_rm(ctx, org, ssh_key_path, force, resource_token):
    docker_client = validate_and_get_docker_client()
    validate_running_resource_manager(ctx.obj, docker_client, force)

    pull_ml_image(ctx.obj, docker_client)
    _validate_apply_config(ctx, docker_client=docker_client, org=org, force=force, ssh_key_path=ssh_key_path, token=resource_token)
    _run_resource_manager(ctx.obj, docker_client)
    click.echo('The resource manager is configured and running')


def cloud_connector_defaults(ctx, cloud_type, kwargs):
    prefix, config_data = get_config_prefix_and_file(ctx.obj)

    return dict(
        mali_image=ctx.obj.mali_image,
        socket_server=ctx.obj.rm_socket_server,
        config_volume=ctx.obj.rm_config_volume,
        rm_image=ctx.obj.rm_manager_image,
        container_name=ctx.obj.rm_container_name,
        prefix=prefix,
        name=pop_key_or_prompt_if(kwargs, 'connector', text='Connector [--connector]:', default='%s-%s' % (cloud_type, 'default')),
        cloud_type=cloud_type,
    ), config_data


def azure_auth(ctx, kwargs, org):
    from .commons import WaitForHttpResponse
    template, _ = cloud_connector_defaults(ctx, cloud_type='azure', kwargs=kwargs)

    azure_subscription_id = pop_key_or_prompt_if(kwargs, 'azure_subscription_id', text='subscription id [--azure-subscription-id]')
    tenant_id_request_url = '{org}/azure/tenant_id_for/{subscription_id}'.format(org=org, subscription_id=azure_subscription_id)

    tenant_id_response = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', tenant_id_request_url)
    tenant_id = tenant_id_response['tenantId']

    server = WaitForHttpResponse(uri='/console/azure/authorised', port=8002)
    cloud_data = [
        {'key': 'auth_state', 'data': 'pending'},
        {'key': 'auth_return_url', 'data': server.url},
        {'key': 'azure_subscription_id', 'data': azure_subscription_id},
        {'key': 'azure_tenant_id', 'data': tenant_id},
    ]
    template['cloud_data'] = cloud_data
    url = '{org}/cloud_connector/{name}'.format(org=org, name=template['name'])
    ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', url, template)

    auth_url = '{org}/azure/authorise_config/{name}'.format(org=org, name=template['name'])
    auth_request = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', auth_url)
    auth_url_path = auth_request['url']
    click.echo(auth_url_path)

    resp = server.run()
    authed_url = '{org}/azure/authorised/{config_id}/code/{code}'.format(
        org=org, config_id=template['name'], code=resp['code'][0])
    authed_request = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', authed_url)
    if authed_request['ok']:
        click.echo('Azure Authorisation for %s completed' % template['name'])
    else:
        click.echo('Azure Authorisation for %s failed' % template['name'])

    return authed_request['result']


def get_cloud_connector_default_connector_name(available_connectors, cloud_type):
    for connector in available_connectors:
        if connector['default']:
            return connector['name']
    if available_connectors:
        return available_connectors[0]['name']
    return '%s-default' % cloud_type


def _get_cloud_connectors(ctx, org, cloud_type):
    click.echo("Checking for available cloud connectors for %s" % cloud_type)
    url = '{org}/{cloud_type}/cloud_connectors'.format(org=org, cloud_type=cloud_type)
    response = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', url)
    return response.get('result', [])


def get_cloud_connector(ctx, org, kwargs, cloud_type, connect_dict=None):
    if connect_dict is None:
        connect_dict = {
            'azure': lambda: azure_auth(ctx, kwargs, org),
        }

    available_connectors = _get_cloud_connectors(ctx, org, cloud_type)
    default_connector = get_cloud_connector_default_connector_name(available_connectors, cloud_type)
    click.echo('List of %s available connectors for %s' % (cloud_type, org))
    output_result(ctx, available_connectors)
    click.echo('You can use easting connector by specifying its name or create a new one by using a name from the list')
    connector_param = 'connector'
    connector_cli_param = '--connector'
    connector_name = pop_key_or_prompt_if(kwargs, connector_param, text='Connector name [%s] ' % connector_cli_param, default=default_connector)
    for x in available_connectors:
        if connector_name == x['name']:
            click.echo('Reusing `%s` connector' % connector_name)
        return x
    kwargs[connector_param] = connector_name

    return connect_dict[cloud_type]()


def _resource_group_defaults(kws, connector, cloud_type):
    is_static = not pop_key_or_prompt_if(kws, 'dynamic_group', type=bool, text=' [--dynamic-group]', default=cloud_type == "aws")
    static_hint = 'static' if is_static else 'dynamic'
    group_name = pop_key_or_prompt_if(kws, 'group_name', text='Group name [--group-name]', default='%s-default' % static_hint)
    click.echo('Creating %s group `%s` ' % (static_hint, group_name))

    return {
        'name': group_name,
        'cloud_type': cloud_type,
        'connector': connector,
        'is_static': is_static,
        'config': '{}',  # todo: add cloud data from the connector here?
        'gpu': pop_key_or_prompt_if(kws, 'group_gpu', text='Is this a GPU group [--group-gpu]', default=False, type=bool),
        'terminate_on_stop': pop_key_or_prompt_if(kws, 'terminate_on_stop', text='Terminate instances (should be %s for %s groups) [--group-terminate-on-stop]' % (not is_static, static_hint), default=not is_static, type=bool),
        'priority': pop_key_or_prompt_if(kws, 'priority', text='Priority (higher is better) [--group-priority]', default=100, type=int),
        'idle_timeout': pop_key_or_prompt_if(kws, 'idle_timeout', text='Idle timeout  [--group-idle-timeout]', default=15, type=int),
    }


def create_group(ctx, kws, org):
    cloud_type = pop_key_or_prompt_if(kws, 'cloud_type', text='Cloud Type [--cloud-type]', default='azure')
    connector = get_cloud_connector(ctx, org, kws, cloud_type)
    request = _resource_group_defaults(kws, connector['name'], connector['cloud_type'])

    if not request['is_static']:
        request['capacity'] = pop_key_or_prompt_if(kws, 'group_capacity', text='Cloud Type [--group-capacity]', default=5, type=int)
    group_name = request['name']
    url = '{org}/resource_group/{name}'.format(org=org, name=group_name)
    response = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', url, request)
    click.echo('Group %s created.' % group_name)
    return response['result']


def select_resource_group(ctx, available_groups, static, kws, org):
    hint_name = '%s-default' % ('dynamic' if not static else 'static')
    available_groups = [x for x in available_groups if x['is_static'] == static]
    if available_groups:
        click.echo('Here is a list of your current groups. You can use one of them or enter a name for new group')
        output_result(ctx, available_groups, fields=['name', 'cloud_type', 'capacity', 'gpu', 'priority'])
        hint_name = available_groups[0]['name']
    group_name = pop_key_or_prompt_if(kws, 'group_name', text='Group name [--group-name]', default=hint_name)
    for group in available_groups:
        if group_name == group['name']:
            click.echo('Selected pre-existing group %s' % group_name)
            return group
    kws['group_name'] = group_name
    kws['dynamic_group'] = False

    return create_group(ctx, kws, org)


def static_resource_groups(ctx, org, kwargs):
    click.echo("Checking for available resource groups for %s" % org)
    url = '{org}/static_resource_groups'.format(org=org)
    response = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', url)
    available_groups = response.get('results', [])
    return select_resource_group(ctx, available_groups, True, kwargs, org)


def get_unmanaged_resources_list(ctx, org, cloud_type, connector):
    url = '{org}/{cloud_type}/unmanaged_machines/{connector}'.format(org=org, cloud_type=cloud_type, connector=connector)
    response = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'get', url)
    unmanaged_resources = response.get('results', [])
    index = 0
    for res in unmanaged_resources:
        index += 1
        res['index'] = index

    return unmanaged_resources


def select_unmanaged_machine(ctx, org, cloud_type, connector, kws):
    available_resources = get_unmanaged_resources_list(ctx, org, cloud_type, connector)
    max_index = len(available_resources)
    output_result(ctx, available_resources, ['index', 'name', 'region', 'resource_group', 'type'])
    if max_index == 0:
        return None

    while True:
        selected = pop_key_or_prompt_if(kws, 'machine_index', text='Machine Index', default=1, type=int)
        if 0 < selected <= max_index:
            return available_resources[selected - 1]


def register_static_unmanaged_machine(ctx, org, group, machine_data):
    data = [{'key': key, 'data': str(data)} for key, data in machine_data.items()]
    url = '{org}/resource_group/{name}/register_static'.format(org=org, name=group['name'])
    result = ctx.obj.handle_api(ctx.obj, ctx.obj.session, 'post', url, dict(data=data))
    return result


@resource_commands.command('add_to_static_cloud_group', help="Add exsisting cloud machine to managed resource group")
@click.pass_context
@click.option('--org', type=str, help='organization to use')
@click.option('--name', type=str, help='configuration name')
@click.option('--group-name', type=str, help='Target group name. Specifying non-existing group will create one')
@click.option('--silent/--no-silent', default=False, help='skip prompts with smart defaults')
def setup_static_template(ctx, org, **kwargs):
    group = static_resource_groups(ctx, org, kwargs)
    machine = select_unmanaged_machine(ctx, org, group['cloud_type'], group['cloud_connector'], kwargs)
    if machine is None:
        click.echo('All of your machines are manged.')
        return
    res = register_static_unmanaged_machine(ctx, org, group, machine)
    click.echo(res)


class RMExpando(object):
    rm_socket_server = None
    rm_manager_image = None
    rm_config_volume = None
    rm_container_name = None


@resource_commands.command('restore_aws_template', help="restores predefined cloud configuration")
@click.pass_context
@click.option('--arn', type=str, help='arn of the KMS encryption key', required=True)
@click.option('--ssh', type=(str, str, str), help='ssh key data', required=True)
@click.option('--mali', type=(str, str, str), help='mali config data', required=True)
@click.option('--prefix', type=str, help='mali prefix type', required=False)
@click.option('--token', type=str, help='mali prefix type', required=True)
@click.option('--rm-socket-server', type=str, help='web socket server', required=True)
@click.option('--rm-manager-image', type=str, required=True)
@click.option('--rm-config-volume', type=str, required=True)
@click.option('--rm-container-name', type=str, required=True)
@click.option('--ml-backend', type=str, required=True)
def apply_aws_template(
        ctx, arn, ssh, mali, prefix, token, rm_socket_server, rm_manager_image,
        rm_config_volume, rm_container_name, ml_backend):
    from .cloud.aws import AwsContext
    if prefix == str(None):
        prefix = None

    click.echo('decrypting data')
    kms = AwsContext.get_kms(arn)
    ssh_key = AwsContext.decrypt(kms, ssh).decode('utf-8')
    mali_data = AwsContext.decrypt(kms, mali).decode('utf-8')
    docker_client = validate_and_get_docker_client()

    click.echo('building installation config')
    config = RMExpando()
    config.rm_socket_server = rm_socket_server
    config.rm_manager_image = rm_manager_image
    config.rm_config_volume = rm_config_volume
    config.rm_container_name = rm_container_name

    click.echo('pulling RM')
    pull_ml_image(ctx.obj, docker_client)

    click.echo('killing RM')
    validate_running_resource_manager(config, docker_client, True)

    click.echo('building volume')
    if _docker_present(docker_client.volumes.get, rm_config_volume):
        docker_client.volumes.get(rm_config_volume).remove(force=True)
    docker_client.volumes.create(rm_config_volume)
    _apply_config_to_volume(config, docker_client, ssh_key, token, prefix=prefix, config_data=mali_data, ml_backend=ml_backend)

    click.echo('Clear containers')
    for container in docker_client.containers.list():
        if container.name == rm_container_name:
            click.echo("\t  KILL: %s" % container.id)
            container.kill()

    for container in docker_client.containers.list(all=True):
        if container.name == rm_container_name:
            click.echo("\t  REMOVE: %s" % container.id)
            container.remove(force=True)

    click.echo('Start RM:')
    inst = _run_resource_manager(config, docker_client)
    click.echo('The resource manager is configured and running %s' % inst.id)
    click.echo('for logs run: docker logs -f %s ' % rm_container_name)



# -*- coding: utf8 -*-
import os


def mkdir_recursive(path):
    sub_path = os.path.dirname(path)
    if not os.path.exists(sub_path):
        mkdir_recursive(sub_path)
    if not os.path.exists(path):
        os.mkdir(path)
